import os
import logging
import inspect
import shlex
import shutil
from argparse import ArgumentParser
from copy import deepcopy
from json import dumps, loads
from typing import List, Dict, Any, Optional

from tqdm import tqdm

from alea.model import StatisticalModel
from alea.runner import Runner
from alea.utils import (
    get_file_path,
    load_yaml,
    compute_variations,
    add_i_batch,
    can_assign_to_typing,
)


class Submitter:
    """Submitter base class that generate the submission script from the configuration. It is
    initialized by the configuration file, and the configuration file should contain the arguments
    of __init__ method of the Submitter.

    Attributes:
        statistical_model (str): the name of the statistical model
        statistical_model_config (str): the configuration file of the statistical model
        poi (str): the parameter of interest
        computation_dict (dict): the dictionary of the computation,
            with keys to_zip, to_vary and in_common
        debug (bool): whether to run in debug mode.
            If True, only one job will be submitted or one runner will be returned.
            And its script will be printed.
        resubmit (bool): whether to resubmit the jobs that have not finished.
            If True, will submit all the jobs, even if the output file exists.

    Args:
        statistical_model (str): the name of the statistical model
        statistical_model_config (str): the configuration file of the statistical model
        poi (str): the parameter of interest
        computation_options (dict): the configuration of the computation
        computation (str, optional (default='discovery_power')): the name of the computation,
            it should be a key of computation_options
        outputfolder (str, optional (default=None)): the output folder
        debug (bool, optional (default=False)): whether to run in debug mode
        loglevel (str, optional (default='INFO')): the log level

    Keyword Args:
        kwargs: the arguments of __init__ method of the Submitter,
            containing configurations of clusters

    Caution:
        All the source of template should be from the same folder.
        All the output, including toydata and fitting results, should be in the same folder.

    """

    config_file_path: str
    template_path: str
    combine_n_jobs: int = 1
    first_i_batch: int = 0
    allowed_special_args: List[str] = []
    logging = logging.getLogger("submitter_logger")

    def __init__(
        self,
        statistical_model: str,
        statistical_model_config: str,
        poi: str,
        computation_options: dict,
        computation: str = "discovery_power",
        outputfolder: Optional[str] = None,
        fit_strategy: Optional[dict] = None,
        debug: bool = False,
        resubmit: bool = False,
        loglevel: str = "INFO",
        **kwargs,
    ):
        """Initializes the submitter."""
        if type(self) == Submitter:  # noqa: E721
            raise RuntimeError(
                "You cannot instantiate the Submitter class directly, "
                "you must use a subclass where the submit method are implemented"
            )
        loglevel = getattr(logging, loglevel.upper() if not debug else "DEBUG")
        self.logging.setLevel(loglevel)

        # find the path of template, requires users install alea-inference properly
        self.run_toymc = shutil.which("alea-run_toymc")
        if self.run_toymc is None:
            raise RuntimeError(
                "Excecutable alea-run_toymc is not found, "
                "please make sure you have installed alea-inference correctly, "
                "and appended alea/bin or .local/bin(pip install direction) to your $PATH."
            )

        self.statistical_model = statistical_model
        self.statistical_model_config = statistical_model_config
        self.poi = poi
        self.outputfolder = outputfolder
        self.fit_strategy = fit_strategy

        self.computation = computation
        self.computation_dict = computation_options[self.computation]
        self.debug = debug
        self.resubmit = resubmit

        # Find statistical model config file
        if not os.path.exists(self.statistical_model_config):
            self.statistical_model_config = os.path.join(
                os.path.dirname(get_file_path(self.config_file_path)), self.statistical_model_config
            )
        if not (
            os.path.exists(self.statistical_model_config)
            and os.path.isfile(self.statistical_model_config)
        ):
            raise FileNotFoundError(
                f"statistical_model_config {self.statistical_model_config} "
                "is not a valid filename or does not exist, "
                "presumably it should be in the same folder as "
                f"config_file_path {self.config_file_path}."
            )

        # Initialize the statistical model
        statistical_model_class = StatisticalModel.get_model_from_name(self.statistical_model)
        self.model = statistical_model_class.from_config(
            self.statistical_model_config, template_path=self.template_path
        )

        # Get fittable and not fittable parameters, for parameters classification later
        self.parameters_fittable = self.model.parameters.fittable + ["poi_expectation"]
        self.parameters_not_fittable = self.model.parameters.not_fittable

    @classmethod
    def from_config(cls, config_file_path: str, **kwargs) -> "Submitter":
        """Initializes the submitter from a yaml config file.

        Args:
            config_file_path (str): Path to the yaml config file.

        Returns:
            BlueiceExtendedModel: Statistical model.

        """
        config = load_yaml(config_file_path)
        cls.config_file_path = config_file_path
        return cls(**{**config, **kwargs})

    @property
    def outputfolder(self) -> Optional[str]:
        return self._outputfolder

    @outputfolder.setter
    def outputfolder(self, outputfolder: Optional[str]):
        if outputfolder is None:
            # default output folder is the current working directory
            raise ValueError("outputfolder is not provided")
        else:
            self._outputfolder = os.path.abspath(outputfolder)
        if not os.path.exists(self._outputfolder):
            os.makedirs(self._outputfolder, exist_ok=True)

    @staticmethod
    def arg_to_str(value, annotation) -> str:
        """Convert the argument to string for the submission script.

        Args:
            value: the value of the argument, can be various type
            annotation: the annotation of the argument

        Returns:
            str: the string of the argument

        Caution:
            Currently we only support str, int, float, bool, dict and list.
            The float will be rounded to 4 digits after the decimal point.

        """
        if value is None:
            return "None"
            # raise ValueError('provides argument can not be None')
        if can_assign_to_typing(str, annotation):
            return value
        elif can_assign_to_typing(int, annotation):
            return "{:d}".format(value)
        elif can_assign_to_typing(float, annotation):
            # currently we only support 4 digits after the decimal point
            rounded_value = round(value, 4)
            if rounded_value != value:
                logging.warn(
                    f"Value {value} of {annotation} will be rounded to four "
                    f"digit precision for the submission script.",
                )
            return "{:.4f}".format(value)
        elif can_assign_to_typing(bool, annotation):
            return str(value)
        elif can_assign_to_typing(dict, annotation) or can_assign_to_typing(list, annotation):
            # the replacement is needed because the json.dumps adds spaces
            return dumps(value).replace(" ", "")
        else:
            raise ValueError(
                f"Unknown annotation type: {annotation}, "
                "it can only be str, int, float, bool, dict or list, "
                "or the typing relatives of them."
            )

    @staticmethod
    def str_to_arg(value: str, annotation):
        """Convert the string to argument for the submission script.

        Args:
            value: the string of the argument
            annotation: the annotation of the argument

        Returns:
            the value of the argument, can be various type

        """
        if value == "None":
            return None
        if can_assign_to_typing(str, annotation):
            return value
        elif can_assign_to_typing(int, annotation):
            return int(value)
        elif can_assign_to_typing(float, annotation):
            return float(value)
        elif can_assign_to_typing(bool, annotation):
            if value == "True":
                return True
            elif value == "False":
                return False
            else:
                raise ValueError(f"Unknown value type: {value}, it can only be True or False")
        elif can_assign_to_typing(dict, annotation) or can_assign_to_typing(list, annotation):
            # the replacement is needed because the json.dumps adds spaces
            return loads(value)
        else:
            raise ValueError(
                f"Unknown annotation type: {annotation}, "
                "it can only be str, int, float, bool, dict or list, "
                "or the typing relatives of them."
            )

    def merged_arguments_generator(self):
        """Generate the merged arguments for Runner from to_zip, to_vary and in_common."""
        _, default_args, _ = Runner.runner_arguments()

        to_zip = self.computation_dict.get("to_zip", {})
        to_vary = self.computation_dict.get("to_vary", {})
        in_common = self.computation_dict.get("in_common", {})
        allowed_keys = ["to_zip", "to_vary", "in_common"]
        if set(self.computation_dict.keys()) - set(allowed_keys):
            raise ValueError(
                "Keys in computation_options should be to_zip, to_vary or in_common, "
                "unknown computation options: {}".format(
                    set(self.computation_dict.keys()) - set(allowed_keys)
                )
            )

        merged_args_list = compute_variations(to_zip=to_zip, to_vary=to_vary, in_common=in_common)

        common_runner_args = {
            "statistical_model": self.statistical_model,
            "statistical_model_config": self.statistical_model_config,
            "poi": self.poi,
            "fit_strategy": self.fit_strategy,
        }

        if set(merged_args_list[0].keys()) & set(common_runner_args.keys()):
            raise ValueError(
                "You specified the following arguments in computation_options, "
                "but they are already specified in the submitter: "
                f"{set(merged_args_list[0].keys()) & set(common_runner_args.keys())}."
            )

        if self.debug:
            print("\n\n" + f"Will submit {len(merged_args_list)} argument combinations:")
            for merged_args in merged_args_list:
                print(merged_args)
        for merged_args in tqdm(merged_args_list):
            runner_args = deepcopy(default_args)
            # update defaults with merged_args and common_runner_args
            runner_args.update(merged_args)
            runner_args.update(common_runner_args)

            # update n_mc if n_batch is provided
            self.update_n_batch(runner_args)
            # update folder and i_batch
            self.update_output_toydata(runner_args, self.outputfolder)
            # update generate_values and nominal_values for runner
            self.update_runner_args(
                runner_args, self.parameters_fittable, self.parameters_not_fittable
            )
            # update the path of limit_threshold
            self.update_limit_threshold(runner_args, self.outputfolder)
            # update template_path and limit_threshold in statistical_model_args if needed
            self.update_statistical_model_args(runner_args, self.template_path)
            # check if all arguments are supported
            self.check_redunant_arguments(runner_args, self.allowed_special_args)

            yield runner_args

    def filename_kwargs(self, runner_args: dict) -> dict:
        """Get the filename_kwargs from runner_args.

        Args:
            runner_args (dict): the arguments of Runner

        Returns:
            dict: the keyword arguments for the filename

        """
        needed_kwargs = {
            "i_batch": runner_args["i_batch"],
            "confidence_level": runner_args["confidence_level"],
            **runner_args["nominal_values"],
            **runner_args["generate_values"],
        }
        return needed_kwargs

    def computation_tickets_generator(self):
        """Get the submission script for the current configuration. It generates the submission
        script for each combination of the computation options.

        For Runner from to_zip, to_vary and in_common:
            - First, generate the combined computational options directly.
            - Second, update the input and output folder of the options.
            - Thrid, collect the non-fittable(settable) parameters into nominal_values.
            - Then, collect the fittable parameters into generate_values.
            - Finally, it generates the submission script for each combination.

        Yields:
            (str, str): the submission script and name output_filename

        """

        _, _, annotations = Runner.runner_arguments()

        for runner_args in self.merged_arguments_generator():
            for i_batch in range(self.first_i_batch, runner_args.get("n_batch", 1)):
                i_args = deepcopy(runner_args)
                i_args["i_batch"] = i_batch

                for name in ["output_filename", "toydata_filename", "limit_threshold"]:
                    if i_args.get(name, None) is not None:
                        # Note: here the later format will overwrite the previous one,
                        # so generate_values have the highest priority.
                        needed_kwargs = self.filename_kwargs(i_args)
                        try:
                            i_args[name] = i_args[name].format(**needed_kwargs)
                        except KeyError:
                            raise KeyError(
                                f"Keys for {i_args[name]} are not in provided arguments "
                                f"{needed_kwargs}, please check the {name}."
                            )

                script = Submitter.script_from_runner_kwargs(annotations, i_args)
                script = f"python3 {self.run_toymc} " + " ".join(
                    map(shlex.quote, script.split(" "))
                )

                if not self.already_done(i_args):
                    yield script, i_args["output_filename"]

    def already_done(self, i_args: dict) -> bool:
        """Check if the job is already done, considering the modes of toydata and output."""
        toydata_mode = i_args["toydata_mode"]
        toydata_filename = i_args["toydata_filename"]
        only_toydata = i_args["only_toydata"]
        output_filename = i_args["output_filename"]
        # these check might need change if we support more modes
        if (toydata_mode == "generate_and_store") and (toydata_filename is None):
            raise ValueError(
                "toydata_filename should be provided when toydata_mode is generate_and_store."
            )
        if (not only_toydata) and (output_filename is None):
            raise ValueError("output_filename should be provided when only_toydata is False.")

        is_done = True
        if self.resubmit or (self.computation == "threshold"):
            is_done = False
        if (toydata_mode == "generate_and_store") and (not os.path.exists(toydata_filename)):
            is_done = False
        if (not only_toydata) and (not os.path.exists(output_filename)):
            is_done = False
        return is_done

    def combined_tickets_generator(self):
        """Get the combined submission script for the current configuration. ``self.combine_n_jobs``
        jobs will be combined into one submission script.

        Yields:
            (str, str): the combined submission script and name output_filename

        Note:
            User can add ``combine_n_jobs: 10`` in ``local_configurations``,
            ``slurm_configurations`` or ``htcondor_configurations`` to combine 10 jobs into
            one submission script. User will need this feature when the number of jobs pending
            for submission is too large.

        """

        _script = ""
        n_combined = 0
        n_submitted = 0
        for script, last_output_filename in self.computation_tickets_generator():
            if n_combined == 0:
                _script += script
            else:
                _script += " && " + script
            n_combined += 1
            if n_combined == self.combine_n_jobs:
                yield _script, last_output_filename
                n_submitted += 1
                n_combined = 0
                _script = ""
        else:
            if n_combined > 0:
                yield _script, last_output_filename
                n_submitted += 1
        print(f"Total {n_submitted} jobs submitted.")

    @staticmethod
    def update_n_batch(runner_args):
        """Update n_mc if n_batch is provided.

        Distribute n_mc into n_batch, so that each batch will run n_mc/n_batch times.

        """
        if "n_mc" not in runner_args:
            logging.warn("n_mc is not provided, it will be set to the default value of Runner")
            return
        if "n_batch" in runner_args:
            if runner_args["n_mc"] % runner_args["n_batch"] != 0:
                raise ValueError("n_mc must be divisible by n_batch")
            runner_args["n_mc"] = runner_args["n_mc"] // runner_args["n_batch"]

    @staticmethod
    def update_output_toydata(runner_args, outputfolder: str):
        for f in ["output_filename", "toydata_filename"]:
            if (f in runner_args) and (runner_args[f] is not None):
                if ("n_batch" in runner_args) and (runner_args["n_batch"] != 1):
                    runner_args[f] = os.path.join(outputfolder, add_i_batch(runner_args[f]))
                else:
                    runner_args[f] = os.path.join(outputfolder, runner_args[f])

    @staticmethod
    def update_runner_args(
        runner_args: Dict[str, Dict[str, Any]],
        parameters_fittable: List[str],
        parameters_not_fittable: List[str],
    ):
        """Update the runner arguments' generate_values and nominal_values. If the argument is
        fittable, it will be added to generate_values, otherwise it will be added to nominal_values.

        Args:
            runner_args (dict): the arguments of Runner

        """
        if runner_args["generate_values"] is None:
            runner_args["generate_values"] = {}
        if runner_args["nominal_values"] is None:
            runner_args["nominal_values"] = {}
        kw_to_pop = []
        for k, v in runner_args.items():
            if k in parameters_fittable:
                runner_args["generate_values"][k] = v
                kw_to_pop.append(k)
            elif k in parameters_not_fittable:
                runner_args["nominal_values"][k] = v
                kw_to_pop.append(k)
        for k in kw_to_pop:
            runner_args.pop(k)
        if set(runner_args["generate_values"].keys()) - set(parameters_fittable):
            raise ValueError(
                f'The generate_values {runner_args["generate_values"]} '
                f"should be a subset of the fittable parameters "
                f"{parameters_fittable} in the statistical model."
            )
        if not all([isinstance(v, (float, int)) for v in runner_args["generate_values"].values()]):
            raise ValueError(
                f"The generate_values {runner_args['generate_values']} "
                "should be all float or int."
            )
        if not all([isinstance(v, (float, int)) for v in runner_args["nominal_values"].values()]):
            raise ValueError(
                f"The nominal_values {runner_args['nominal_values']} should be all float or int."
            )

    @staticmethod
    def update_limit_threshold(runner_args, outputfolder: str):
        if "limit_threshold" in runner_args:
            runner_args["limit_threshold"] = os.path.join(
                outputfolder, runner_args["limit_threshold"]
            )

    @staticmethod
    def update_statistical_model_args(
        runner_args: Dict[str, Dict[str, Any]], template_path: Optional[str] = None
    ):
        """Update template_path in the statistical model arguments.

        Args:
            runner_args (dict): the arguments of Runner

        """
        if runner_args["statistical_model_args"] is None:
            runner_args["statistical_model_args"] = {}
        if template_path is not None:
            runner_args["statistical_model_args"]["template_path"] = template_path
        if "limit_threshold" in runner_args:
            runner_args["statistical_model_args"]["limit_threshold"] = runner_args.pop(
                "limit_threshold"
            )
        if "limit_threshold_interpolation" in runner_args:
            runner_args["statistical_model_args"]["limit_threshold_interpolation"] = (
                runner_args.pop("limit_threshold_interpolation")
            )
        if "asymptotic_dof" in runner_args:
            runner_args["statistical_model_args"]["asymptotic_dof"] = runner_args.pop(
                "asymptotic_dof"
            )

    @staticmethod
    def check_redunant_arguments(runner_args, allowed_special_args: List[str] = []):
        signatures = inspect.signature(Runner.__init__)
        args = list(signatures.parameters.keys())[1:] + ["n_batch"] + allowed_special_args
        intended_args = set(runner_args.keys())
        allowed_args = set(args)
        if not intended_args.issubset(allowed_args):
            raise ValueError(
                f"Not all arguments are supported, "
                f"arguments {allowed_args} are acceptable, "
                f"and the following arguments is unknown: "
                f"{intended_args - allowed_args}."
            )

    def submit(self, *arg, **kwargs):
        """Submit the jobs to the destinations."""
        raise NotImplementedError("You must write a submit function your submitter class")

    def all_runner_kwargs(self):
        """Parse all the runner arguments from the submission script."""
        kwargs_list = []
        for _, (script, _) in enumerate(self.computation_tickets_generator()):
            kwargs = Submitter.runner_kwargs_from_script(shlex.split(script)[2:])
            kwargs_list.append(kwargs)
        return kwargs_list

    @staticmethod
    def runner_kwargs_from_script(sys_argv: Optional[List[str]] = None):
        """Parse kwargs of a Runner from a string of arguments(script).

        Args:
            sys_argv (list, optional (default=None)): string of arguments, with the format of
                ['--arg1', 'value1', '--arg2', 'value2', ...]. The arguments must be the same as
                the arguments of Runner.__init__.

        """
        signatures = inspect.signature(Runner.__init__)
        args = list(signatures.parameters.keys())[1:]
        parser = ArgumentParser(description="Command line running of alea-run_toymc")

        # skip the first one because it is self(Runner itself)
        for arg in args:
            parser.add_argument(f"--{arg}", type=str, required=True, help=None)

        parsed_args = parser.parse_args(args=sys_argv)
        kwargs = {}
        for arg, value in parsed_args.__dict__.items():
            kwargs.update({arg: Submitter.str_to_arg(value, signatures.parameters[arg].annotation)})
        return kwargs

    @staticmethod
    def script_from_runner_kwargs(annotations, kwargs) -> str:
        """Generate the submission script from the runner arguments."""
        script_array = []
        for arg, annotation in annotations.items():
            script_array.append(f"--{arg}")
            script_array.append(Submitter.arg_to_str(kwargs[arg], annotation))
        script = " ".join(script_array)
        return script



import shutil

import pytest
from unittest import TestCase

from alea.submitter import Submitter
from alea.submitters.local import SubmitterLocal


@pytest.mark.usefixtures("rm_cache")
class TestSubmitter(TestCase):
    """Test of the Submitter class."""

    @classmethod
    def setUp(cls):
        """Initialise the Submitter instance."""
        cls.config_file_path = "unbinned_wimp_running.yaml"
        cls.outputfolder = "test_output"
        cls.set_submitter(cls)

    def set_submitter(self):
        """Set a new submitter instance with SubmitterLocal."""
        self.submitter = SubmitterLocal.from_config(
            self.config_file_path,
            computation="discovery_power",
            debug=True,
            outputfolder=self.outputfolder,
        )

    def test_init(self):
        """Test of __init__ method."""
        with self.assertRaises(
            RuntimeError, msg="Should raise error when directly instantiating Submitter"
        ):
            Submitter.from_config(self.config_file_path)

    def test_submit(self):
        """Test of submit method."""
        self.submitter.submit()
        shutil.rmtree(self.outputfolder, ignore_errors=False)

    def test_all_runner_kwargs(self):
        """Test of all_runner_kwargs method."""
        self.submitter.all_runner_kwargs()



__version__ = "0.2.8"

from .parameters import *

from .simulators import *

from .template_source import *

from .model import *

from .models import *

from .runner import *

from .submitter import *

from .utils import *

from .examples import *



import pytest
import shutil


@pytest.fixture(scope="class")
def rm_cache():
    """Remove pdf_cache directory before and after initializing the TestCase."""
    shutil.rmtree("pdf_cache", ignore_errors=True)
    yield
    shutil.rmtree("pdf_cache", ignore_errors=True)



import logging
from typing import List, Dict, Optional, Union

import numpy as np
from scipy.interpolate import interp1d
from blueice import HistogramPdfSource
from multihist import Hist1d
from inference_interface import template_to_multihist

from alea.utils import load_json

logging.basicConfig(level=logging.INFO)
can_check_binning = True


class TemplateSource(HistogramPdfSource):
    """A source defined with a template histogram. The parameters are set in self.config.
    "templatename", "histname", "analysis_space" must be in self.config.

    Attributes:
        config (dict): The configuration of the source.
        dtype (list): The data type of the source.
        _bin_volumes (numpy.ndarray): The bin volumes of the source.
        _n_events_histogram (multihist.MultiHistBase):
            The histogram of the number of events of the source.
        events_per_day (float): The number of events per day of the source.
        _pdf_histogram (multihist.MultiHistBase):
            The histogram of the probability density function of the source.

    Args:
        config (dict): The configuration of the source.
        templatename: Hdf5 file to open.
        histname: Histogram name.
        named_parameters (list):
            List of config setting names to pass to .format on histname and filename.
        normalise_template (bool): Normalise the template histogram.
        in_events_per_bin (bool):
            If True, histogram is in events per day / bin.
            If False or absent, histogram is already pdf.
        histogram_scale_factor (float): Multiply histogram by this number
        convert_to_uniform (bool): Convert the histogram to a uniform per bin distribution.
        log10_bins (list):
            List of axis numbers.
            If True, bin edges on this axis in the hdf5 file are log10() of the actual bin edges.

    """

    def __init__(self, config: Dict, *args, **kwargs):
        """Initialize the TemplateSource."""
        # override the default interpolation method
        if "pdf_interpolation_method" not in config:
            config["pdf_interpolation_method"] = "piecewise"
        super().__init__(config, *args, **kwargs)

    def _check_binning(self, h, histogram_info: str):
        """Check if the histogram"s bin edges are the same to analysis_space.

        Args:
            h (multihist.MultiHistBase): The histogram to check.
            histogram_info (str): Information of the histogram.

        """
        # Deal with people who have log10"d their bins
        for axis_i in self.config.get("log10_bins", []):
            h.bin_edges[axis_i] = 10 ** h.bin_edges[axis_i]

        analysis_space = self.config["analysis_space"]
        if len(analysis_space) != h.histogram.ndim:
            raise ValueError(
                f"Analysis space {analysis_space} has {len(analysis_space)} axes, "
                f"but histogram {histogram_info} has {h.histogram.ndim} axes."
            )

        # Check if the histogram bin edges are correct
        for axis_i, (_, expected_bin_edges) in enumerate(analysis_space):
            expected_bin_edges = np.array(expected_bin_edges)
            # If 1D, hist1d returns bin_edges straight, not as list
            if isinstance(h, Hist1d):
                seen_bin_edges = h.bin_edges
            else:
                seen_bin_edges = h.bin_edges[axis_i]
            logging.debug("axis_i: " + str(axis_i))
            logging.debug("expected_bin_edges: " + str(expected_bin_edges))
            logging.debug("seen_bin_edges: " + str(seen_bin_edges))
            logging.debug("h.bin_edges type" + str(h.bin_edges))
            err_msg = (
                f"Axis {axis_i:d} of histogram {histogram_info} "
                f"has bin edges {seen_bin_edges}, but expected {expected_bin_edges}."
            )
            if len(seen_bin_edges) != len(expected_bin_edges):
                raise ValueError(err_msg)
            np.testing.assert_almost_equal(
                seen_bin_edges, expected_bin_edges, decimal=2, err_msg=err_msg
            )

    @property
    def format_named_parameters(self):
        """Get the named parameters in the config to dictionary format."""
        format_named_parameters = {
            k: self.config[k] for k in self.config.get("named_parameters", [])
        }
        return format_named_parameters

    def build_histogram(self):
        """Build the histogram of the source."""
        templatename = self.config["templatename"].format(**self.format_named_parameters)
        histname = self.config["histname"].format(**self.format_named_parameters)
        h = template_to_multihist(templatename, histname)
        if np.min(h.histogram) < 0:
            raise AssertionError(f"There are bins for source {templatename} with negative entries.")

        if self.config.get("normalise_template", False):
            h /= h.n
        if not self.config.get("in_events_per_bin", True):
            h.histogram *= h.bin_volumes()

        h = self.apply_slice_args(h)

        # Fix the bin sizes
        if can_check_binning:
            histogram_info = f"{histname:s} in hdf5 file {templatename:s}"
            self._check_binning(h, histogram_info)

        self.set_dtype()
        self.set_pdf_histogram(h)

    def apply_slice_args(self, h, slice_args: Optional[Union[List[Dict], Dict]] = None):
        """Apply slice arguments to the histogram.

        Args:
            h (multihist.MultiHistBase): The histogram to apply the slice arguments to.
            slice_args (dict): The slice arguments to apply.
                The sum_axis, slice_axis, and slice_axis_limits are supported.

        """

        # if slice_args is not specified, use the one in the config
        if slice_args is None:
            slice_args = self.config.get("slice_args", [{}])
        # if slice_args is a dict, convert it to a list
        if isinstance(slice_args, dict):
            slice_args = [slice_args]
        # if slice_args is empty, return
        if (slice_args is None) or (len(slice_args) == 0) or (slice_args == [{}]):
            return h
        # check if slice_args is a list of dicts, and if each dict has slice_axis
        if not isinstance(slice_args, list):
            raise ValueError("slice_args must be a list or a dict")
        if any(not isinstance(sa, dict) for sa in slice_args):
            raise ValueError("slice_args must be a list of dicts")
        if any("slice_axis" not in sa for sa in slice_args):
            raise ValueError("slice_axis must be specified in slice_args")

        for sa in slice_args:
            # read slice_axis, sum_axis, and slice_axis_limits from slice_args
            slice_axis = sa["slice_axis"]
            sum_axis = sa.get("sum_axis", False)
            bin_edges = h.bin_edges[h.get_axis_number(slice_axis)]
            slice_axis_limits = sa.get("slice_axis_limits", [bin_edges[0], bin_edges[-1]])

            # slice and/or sum over axis
            if sum_axis:
                logging.debug(
                    f"Slice and sum over axis {slice_axis} from {slice_axis_limits[0]} "
                    f"to {slice_axis_limits[1]}"
                )
                h = h.slicesum(
                    start=slice_axis_limits[0], stop=slice_axis_limits[1], axis=slice_axis
                )
            else:
                logging.debug(
                    f"Slice over axis {slice_axis} from {slice_axis_limits[0]} "
                    f"to {slice_axis_limits[1]}"
                )
                logging.debug(f"Normalization before slicing: {h.n}.")
                h = h.slice(start=slice_axis_limits[0], stop=slice_axis_limits[1], axis=slice_axis)
                logging.debug(f"Normalization after slicing: {h.n}.")

        if isinstance(h, Hist1d):
            if self.config["pdf_interpolation_method"] != "piecewise":
                raise ValueError(
                    "pdf_interpolation_method must be piecewise for 1D histograms. "
                    "This is required by blueice."
                )
        return h

    def set_dtype(self):
        """Set the data type of the source."""
        self.dtype = []
        for n, _ in self.config["analysis_space"]:
            self.dtype.append((n, float))
        self.dtype.append(("source", int))

    def set_pdf_histogram(self, h):
        """Set the histogram of the probability density function of the source."""
        self._bin_volumes = h.bin_volumes()
        # Should not be in HistogramSource... anyway
        self._n_events_histogram = h.similar_blank_histogram()

        histogram_scale_factor = self.config.get("histogram_scale_factor", 1)
        h *= histogram_scale_factor
        logging.debug(
            f"Multiplying histogram with histogram_scale_factor {histogram_scale_factor}. "
            f"Histogram is now normalised to {h.n}."
        )

        self.events_per_day = h.n
        logging.debug(f"events_per_day: {self.events_per_day}")

        # Convert h to density...
        if self.config.get("in_events_per_bin", True):
            h.histogram /= self._bin_volumes

        # ... and finally to probability density
        if self.events_per_day > 0:
            h.histogram /= self.events_per_day
        else:
            logging.warn("Sum of histogram is zero.")

        if self.config.get("convert_to_uniform", False):
            h.histogram = 1.0 / self._bin_volumes

        self._pdf_histogram = h
        logging.debug(f"Setting _pdf_histogram normalised to {h.n}.")

    def simulate(self, n_events: int):
        """Simulate events from the source.

        Args:
            n_events (int): The number of events to simulate.

        Returns:
            numpy.ndarray: The simulated events.

        """
        ret = np.zeros(n_events, dtype=self.dtype)
        h = self._pdf_histogram * self._bin_volumes
        t = h.get_random(n_events)
        for i, (n, _) in enumerate(self.config["analysis_space"]):
            ret[n] = t[:, i]
        return ret


class CombinedSource(TemplateSource):
    """Source that is a weighted sums of histograms. Useful e.g. for safeguard. The first histogram
    is the base histogram, and the rest are added to it with weights. The weights can be set as
    shape parameters in the config.

    Args:
        weights: Weights of the 2nd to the last histograms.
        histnames: List of filenames containing the histograms.
        templatenames: List of names of histograms within the hdf5 files.

    """

    def build_histogram(self):
        """Build the histogram of the source."""
        if not self.config.get("in_events_per_bin", True):
            raise ValueError("CombinedSource does not support in_events_per_bin=False")
        # Check if all the necessary parameters, like weights are specified
        if "weight_names" not in self.config:
            raise ValueError("weight_names must be specified")
        find_weight = [weight_name in self.config for weight_name in self.config["weight_names"]]
        if not all(find_weight):
            missing_weight = self.config["weight_names"][find_weight]
            raise ValueError(f"All weight_names must be specified, but {missing_weight} are not. ")
        weights = [self.config[weight_name] for weight_name in self.config["weight_names"]]
        histnames = self.config["histnames"]
        templatenames = self.config["templatenames"]
        if len(histnames) != len(templatenames):
            raise ValueError("histnames and templatenames must be the same length")
        if len(weights) + 1 != len(templatenames):
            raise ValueError("weights must be 1 shorter than histnames, templatenames")

        slice_args = self.config.get("slice_args", {})
        if isinstance(slice_args, dict):
            slice_args = [slice_args]
        if all(isinstance(sa, dict) for sa in slice_args):
            # Recognize as a single slice_args for all histograms
            slice_argss = [slice_args] * len(histnames)
        else:
            # Recognize as a list of slice_args for each histogram
            slice_argss = slice_args

        histograms = []
        slice_fractions = []
        for histname, templatename, slice_args in zip(histnames, templatenames, slice_argss):
            templatename = templatename.format(**self.format_named_parameters)
            histname = histname.format(**self.format_named_parameters)
            h = template_to_multihist(templatename, histname)
            slice_fraction = 1.0 / h.n
            if h.n == 0.0:
                slice_fraction = 0.0
            h = self.apply_slice_args(h, slice_args)
            slice_fraction *= h.n
            slice_fractions.append(slice_fraction)

            # Avoid dividing by zero for empty PDFs (if mu=0, does not matter if unnormalised)
            if h.n <= 0.0:
                h *= 0.0
            else:
                h /= h.n
            histograms.append(h)
        logging.debug("slice fractions are", slice_fractions)

        for i in range(len(weights)):
            h_comp = histograms[0].similar_blank_histogram()
            h = histograms[i + 1]
            hslices = []
            for j, bincs in enumerate(h.bin_centers()):
                hsliced = h_comp.get_axis_bin_index(bincs[0], j)
                hsliceu = h_comp.get_axis_bin_index(bincs[-1], j) + 1
                hslices.append(slice(hsliced, hsliceu, 1))
            h_comp[tuple(hslices)] += h.histogram
            histograms[0] += h_comp * weights[i]
        h = histograms[0]

        logging.debug("Setting zero for NaNs and Infs in combined template.")
        h.histogram[np.isinf(h.histogram)] = 0.0
        h.histogram[np.isnan(h.histogram)] = 0.0
        h.histogram[h.histogram < 0] = 0.0

        # Set pdf values that are below 0 to zero:
        if np.min(h.histogram) < 0:
            raise AssertionError(f"There are bins for source {templatename} with negative entries.")
        # Fix the bin sizes
        if can_check_binning:
            histogram_info = f"combined template {histnames} in hdf5 file {templatenames}"
            self._check_binning(h, histogram_info)

        self.set_dtype()
        self.set_pdf_histogram(h)


class SpectrumTemplateSource(TemplateSource):
    """Reweighted template source by 1D spectrum. The first axis of the template is assumed to be
    reweighted.

    Args:
        spectrum_name:
            Name of bbf json-like spectrum file

    """

    @staticmethod
    def _get_json_spectrum(filename: str):
        """Translates bbf-style JSON files to spectra.

        Args:
            filename (str): Name of the JSON file.

        Todo:
            Define the format of the JSON file clearly.

        """
        contents = load_json(filename)
        if "description" in contents:
            logging.debug(contents["description"])
        if "coordinate_system" not in contents:
            raise ValueError("Coordinate system not in JSON file.")
        if "map" not in contents:
            raise ValueError("Map not in JSON file.")
        ret = interp1d(
            contents["coordinate_system"], contents["map"], bounds_error=False, fill_value=0.0
        )
        return ret

    def build_histogram(self):
        """Build the histogram of the source."""
        templatename = self.config["templatename"].format(**self.format_named_parameters)
        histname = self.config["histname"].format(**self.format_named_parameters)
        h = template_to_multihist(templatename, histname)

        if "spectrum_name" not in self.config:
            raise ValueError("spectrum_name not in config")
        spectrum = self._get_json_spectrum(
            self.config["spectrum_name"].format(**self.format_named_parameters)
        )

        # Perform scaling, the first axis is assumed to be reweighted
        # The spectrum is assumed to be probability density (in per the unit of first axis).
        axis = 0
        # h = h.normalize(axis=axis)
        bin_edges = h.bin_edges[axis]
        bin_centers = h.bin_centers(axis=axis)
        slices = [None] * h.histogram.ndim
        slices[axis] = slice(None)
        h.histogram = h.histogram * (spectrum(bin_centers) * np.diff(bin_edges))[tuple(slices)]

        if np.min(h.histogram) < 0:
            raise AssertionError(f"There are bins for source {templatename} with negative entries.")

        if self.config.get("normalise_template", False):
            h /= h.n
        if not self.config.get("in_events_per_bin", True):
            h.histogram *= h.bin_volumes()

        h = self.apply_slice_args(h)

        # Fix the bin sizes
        if can_check_binning:
            histogram_info = f"reweighted {histname:s} in hdf5 file {templatename:s}"
            self._check_binning(h, histogram_info)

        self.set_dtype()
        self.set_pdf_histogram(h)



from typing import Dict, List, Optional, Union

import numpy as np
from scipy import stats

from alea.model import StatisticalModel


class GaussianModel(StatisticalModel):
    """A model of a gaussian measurement, where the model has parameters mu and sigma. For
    illustration, we show how required nominal parameters can be added to the init sigma is fixed in
    this example.

    Args:
        parameter_definition (dict or list, optional (default=None)):
            definition of the parameters of the model

    Caution:
        You must define the nominal values of the parameters (mu, sigma)
        in the parameters definition.

    """

    def __init__(
        self,
        parameter_definition: Optional[Union[Dict[str, dict], List[str]]] = None,
        **kwargs,
    ):
        """Initialise a gaussian model."""
        if parameter_definition is None:
            parameter_definition = ["mu", "sigma"]
        super().__init__(parameter_definition=parameter_definition, **kwargs)

    def _ll(self, mu=None, sigma=None):
        """Log-likelihood of the model.

        Args:
            mu (float, optional (default=None)): mean of the gaussian,
                if None, the nominal value is used
            sigma (float, optional (default=None)): standard deviation of the gaussian,
                if None, the nominal value is used

        """
        hat_mu = self.data[0]["hat_mu"][0]
        return stats.norm.logpdf(x=hat_mu, loc=mu, scale=sigma)

    def _generate_data(self, mu=None, sigma=None):
        """Generate data from the model.

        Args:
            mu (float, optional (default=None)): mean of the gaussian,
                if None, the nominal value is used
            sigma (float, optional (default=None)): standard deviation of the gaussian,
                if None, the nominal value is used

        Returns:
            list: data generated from the model

        """
        hat_mu = stats.norm(loc=mu, scale=sigma).rvs()
        data = [np.array([(hat_mu,)], dtype=[("hat_mu", float)])]
        return data



from unittest import TestCase

import numpy as np
import inference_interface as ii
import multihist as mh
from scipy.stats import chi2

from alea.utils import (
    MAX_FLOAT,
    get_analysis_space,
    formatted_to_asterisked,
    asymptotic_critical_value,
    within_limits,
    clip_limits,
    can_expand_grid,
    expand_grid_dict,
    convert_to_vary,
    deterministic_hash,
    get_file_path,
    signal_multiplier_estimator,
)


class TestUtils(TestCase):
    """Test of the alea.utils."""

    def test_get_analysis_space(self):
        """Test of the get_analysis_space function."""
        analysis_space = [
            {"a": "np.arange(3)"},
            {"b": [0, 1, 2]},
            {"c": "0, 1, 2"},
            {"d": "0 1 2"},
        ]
        get_analysis_space(analysis_space)

    def test_formatted_to_asterisked(self):
        """Test of the formatted_to_asterisked function."""
        self.assertEqual(formatted_to_asterisked("a_{a:.2f}_b_{b:d}"), "a_*_b_*")
        self.assertEqual(formatted_to_asterisked("a_{a:.2f}_b_{b:d}", wildcards="a"), "a_*_b_{b:d}")

    def test_asymptotic_critical_value(self):
        """Test of the asymptotic_critical_value function."""
        confidence_level = 0.9
        critical_value = chi2(1).isf(2 * (1.0 - confidence_level))
        self.assertEqual(asymptotic_critical_value("lower", confidence_level), critical_value)
        self.assertEqual(asymptotic_critical_value("upper", confidence_level), critical_value)
        with self.assertRaises(ValueError):
            asymptotic_critical_value("invalid", confidence_level)
        with self.assertRaises(ValueError):
            asymptotic_critical_value("lower", confidence_level, 2)

        critical_value_central = chi2(2).isf((1.0 - confidence_level))
        self.assertEqual(
            asymptotic_critical_value("central", confidence_level, 2), critical_value_central
        )

    def test_within_limits(self):
        """Test of the within_limits function."""
        self.assertTrue(within_limits(MAX_FLOAT, clip_limits(None)))
        self.assertTrue(within_limits(MAX_FLOAT, clip_limits([0, None])))
        self.assertTrue(within_limits(MAX_FLOAT, clip_limits([None, MAX_FLOAT])))

    def test_can_expand_grid(self):
        """Test of the can_expand_grid function."""
        self.assertTrue(can_expand_grid({"a": [1, 2], "b": [3, 4]}))

    def test_expand_grid_dict(self):
        """Test of the expand_grid_dict function."""
        self.assertEqual(
            expand_grid_dict(["free", {"a": 1, "b": 3}, {"a": [1, 2], "b": [3, 4]}]),
            [
                "free",
                {"a": 1, "b": 3},
                {"a": 1, "b": 3},
                {"a": 1, "b": 4},
                {"a": 2, "b": 3},
                {"a": 2, "b": 4},
            ],
        )

    def test_convert_to_vary(self):
        """Test of the convert_to_zip function."""
        varied = convert_to_vary({"a": [1, 2], "b": [{"c": 3}, {"c": 4}]})
        self.assertNotEqual(id(varied[0]["b"]), id(varied[2]["b"]))
        self.assertNotEqual(id(varied[1]["b"]), id(varied[3]["b"]))
        self.assertEqual(
            varied,
            [
                {"a": 1, "b": {"c": 3}},
                {"a": 1, "b": {"c": 4}},
                {"a": 2, "b": {"c": 3}},
                {"a": 2, "b": {"c": 4}},
            ],
        )

    def test_deterministic_hash(self):
        """Test of the deterministic_hash function."""
        self.assertEqual(deterministic_hash([0, 1]), "si3ifpvg2u")
        self.assertEqual(deterministic_hash(np.array([0, 1])), "si3ifpvg2u")
        self.assertEqual(deterministic_hash({"a": 1, "b": 2}), "shhkapn4q7")
        self.assertEqual(
            deterministic_hash({"a": np.array([0, 1]), "b": np.array([0, 1])}), "anxefavaju"
        )

    def test_signal_multiplier_estimator(self):
        bkg = ii.template_to_multihist(
            get_file_path("er_template_0.ii.h5"), hist_name="er_template"
        )
        sig = ii.template_to_multihist(
            get_file_path("wimp50gev_template.ii.h5"), hist_name="wimp_template"
        )
        data = (bkg + sig * 1e-1).get_random(size=np.random.poisson(bkg.n))
        data = mh.Histdd(*data.T, bins=bkg.bin_edges)
        signal_multiplier_estimator(sig.histogram, bkg.histogram, data.histogram, diagnostic=True)



import os
import time
import inspect

from utilix import batchq

from alea.submitter import Submitter


# suggested default arguments for utilix.batchq.submit_job
# for XENONnT collaboration
BATCHQ_DEFAULT_ARGUMENTS = {
    "hours": 1,  # in the unit of hours
    "mem_per_cpu": 2000,  # in the unit of Mb
    "container": "xenonnt-development.simg",
    "partition": "xenon1t",
    "qos": "xenon1t",
    "cpus_per_task": 1,
    "exclude_nodes": "dali[028-030],midway2-0048",
}


class SubmitterSlurm(Submitter):
    """Submitter for slurm cluster, using utilix.batchq.submit_job. The default batchq arguments are
    defined in BATCHQ_DEFAULT_ARGUMENTS. You can also overwrite them by passing them inside
    configuration file.

    Keyword Args:
        slurm_configurations (dict): The configurations for utilix.batchq.submit_job.
            There can be template_path inside it, indicating the path to the template.

    """

    max_jobs = 100

    def __init__(self, *args, **kwargs):
        """Initialize the SubmitterSlurm class."""
        self.name = self.__class__.__name__
        self.slurm_configurations = kwargs.get("slurm_configurations", {})
        self.template_path = self.slurm_configurations.pop("template_path", None)
        self.combine_n_jobs = self.slurm_configurations.pop("combine_n_jobs", 1)
        self.batchq_arguments = {**BATCHQ_DEFAULT_ARGUMENTS, **self.slurm_configurations}
        self._check_batchq_arguments()
        super().__init__(*args, **kwargs)
        self.log_dir = self.outputfolder

    def _submit(self, job, **kwargs):
        """Submits job to batch queue which actually runs the analysis.

        Args:
            job (str): The job script to be submitted.

        Keyword Args:
            jobname (str): The name of the job.
            log (str): The path to the log file.

        """
        jobname = kwargs.pop("jobname", None)
        if jobname is None:
            jobname = self.name

        log = kwargs.pop("log", None)
        if log is None:
            log = os.path.join(self.log_dir, f"{jobname.lower()}.log")

        kwargs_to_pop = []
        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)
                kwargs_to_pop.append(key)
        for kw in kwargs_to_pop:
            kwargs.pop(kw)

        self.logging.debug(f"Submitting the following job: '{job}'")
        batchq.submit_job(job, jobname=jobname, log=log, **{**self.batchq_arguments, **kwargs})

    def _check_batchq_arguments(self):
        """Check if the self.batchq_arguments are valid."""
        args = set(inspect.signature(batchq.submit_job).parameters)
        if not set(self.batchq_arguments).issubset(args):
            raise ValueError(
                f"The following arguments are not supported of utilix.batchq.submit_job: "
                f"{set(self.batchq_arguments) - set(args)}."
            )

    def submit(self, **kwargs):
        """Submits job to batch queue which actually runs the analysis. Overwrite the
        BATCHQ_DEFAULT_ARGUMENTS by configuration file. If debug is True, only submit the first job.

        Keyword Args:
            jobname (str): The name of the job.

        """
        _jobname = kwargs.pop("jobname", self.name.lower())
        batchq_kwargs = {}
        for job, (script, last_output_filename) in enumerate(self.combined_tickets_generator()):
            if self.debug:
                print(script)
                if job > 0:
                    break
            while batchq.count_jobs(_jobname) > self.max_jobs:
                self.logging.info("Too many jobs. Sleeping for 30s.")
                time.sleep(30)
            batchq_kwargs["jobname"] = f"{_jobname}_{job:03d}"
            if last_output_filename is not None:
                batchq_kwargs["log"] = os.path.join(self.log_dir, f"{last_output_filename}.log")
            self.logging.debug(f"Call '_submit' with job: {job} and kwargs: {batchq_kwargs}.")
            self._submit(script, **batchq_kwargs)



import os
import re
import json
import yaml
import importlib_resources
import itertools
import blueice
from glob import glob
from copy import deepcopy
from pydoc import locate
import logging
from hashlib import sha256
from base64 import b32encode
from collections.abc import Mapping
from typing import Any, List, Dict, Tuple, Optional, Union, cast, get_args, get_origin
from blueice.pdf_morphers import Morpher
from itertools import product

import h5py
import matplotlib.pyplot as plt

# These imports are needed to evaluate strings
import numpy  # noqa: F401
import numpy as np  # noqa: F401
from scipy import stats  # noqa: F401
from scipy.stats import chi2

logging.basicConfig(level=logging.INFO)


MAX_FLOAT = np.sqrt(np.finfo(np.float32).max)


class ReadOnlyDict:
    """A read-only dict."""

    def __init__(self, data):
        self._data = dict(data)

    def __getitem__(self, key):
        return self._data[key]

    def __contains__(self, key):
        return key in self._data

    def __repr__(self):
        return repr(self._data)

    def get(self, key, default=None):
        return self._data.get(key, default)

    # Prevent changes
    def __setitem__(self, key, value):
        raise TypeError(
            "This dictionary is read-only, please initialize a new one in order to change it."
        )

    def __delitem__(self, key):
        raise TypeError(
            "This dictionary is read-only, please initialize a new one in order to change it."
        )

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()


def evaluate_numpy_scipy_expression(value: str):
    """Evaluate numpy(np) and scipy.stats expression."""
    if value.startswith("stats."):
        return eval(value)
    elif value.startswith("np.") or value.startswith("numpy."):
        return eval(value)
    else:
        raise ValueError(f"Expression {value} not understood.")


def evaluate_numpy_scipy_expression_in_dict(d: dict):
    """Evaluate numpy(np) and scipy.stats expression in a dict.

    Example:
        >>> evaluate_numpy_scipy_expression_in_dict({'a': 'np.arange(0, 2, 1)', 'b': [0, 1]})
        {'a': [0, 1], 'b': [0, 1]}

    """
    d_copy = deepcopy(d)
    for k, v in d_copy.items():
        if isinstance(v, str):
            d_copy[k] = evaluate_numpy_scipy_expression(v).tolist()
    return d_copy


def get_analysis_space(analysis_space: list) -> list:
    """Convert analysis_space to a list of tuples with evaluated values."""
    eval_analysis_space = []

    for element in analysis_space:
        for key, value in element.items():
            if isinstance(value, str) and value.startswith("np."):
                eval_element = (key, evaluate_numpy_scipy_expression(value))
            elif isinstance(value, str):
                if "," in value:
                    eval_element = (key, np.fromstring(value, dtype=float, sep=","))
                else:
                    eval_element = (key, np.fromstring(value, dtype=float, sep=" "))
            elif isinstance(value, list):
                eval_element = (key, np.array(value))
            else:
                raise ValueError(f"analysis_space for dimension {key} not understood.")

            eval_analysis_space.append(eval_element)
    return eval_analysis_space


def _prefix_file_path(
    config: dict, template_folder_list: list, ignore_keys: List[str] = ["name", "histname"]
):
    """Prefix file path with template_folder_list whenever possible.

    Args:
        config (dict): dictionary contains file path
        template_folder_list (list): list of possible base folders. Ordered by priority.
        ignore_keys (list, optional (default=["name", "histname"])):
        keys to be ignored when prefixing

    """
    for key in config.keys():
        if isinstance(config[key], str) and key not in ignore_keys:
            try:
                config[key] = get_file_path(config[key], template_folder_list)
            except RuntimeError:
                pass


def adapt_likelihood_config_for_blueice(
    likelihood_config: dict, template_folder_list: list
) -> dict:
    """Adapt likelihood config to be compatible with blueice.

    Args:
        likelihood_config (dict): likelihood config dict
        template_folder_list (list): list of possible base folders. Ordered by priority.

    Returns:
        dict: adapted likelihood config

    """

    likelihood_config_copy = deepcopy(likelihood_config)

    likelihood_config_copy["analysis_space"] = get_analysis_space(
        likelihood_config_copy["analysis_space"]
    )

    _prefix_file_path(likelihood_config_copy, template_folder_list)

    if "default_source_class" in likelihood_config_copy:
        default_source_class = locate(likelihood_config_copy["default_source_class"])
        if default_source_class is None:
            raise ValueError(f"Could not find {likelihood_config_copy['default_source_class']}!")
        likelihood_config_copy["default_source_class"] = default_source_class

    for source in likelihood_config_copy["sources"]:
        if "template_filename" in source:
            source["templatename"] = get_file_path(
                source["template_filename"], template_folder_list
            )
        if "class" in source:
            source_class = locate(source["class"])
            if source_class is None:
                raise ValueError(f"Could not find {source['class']}!")
            source["class"] = source_class
        if "template_filenames" in source:
            source["templatenames"] = [
                get_file_path(template_filename, template_folder_list)
                for template_filename in source["template_filenames"]
            ]
        if source.get("efficiency_name", None):
            source["apply_efficiency"] = True
        _prefix_file_path(source, template_folder_list)
    return likelihood_config_copy


def load_yaml(file_name: str):
    """Load data from yaml file."""
    with open(get_file_path(file_name), "r") as file:
        data = yaml.safe_load(file)
    return data


def load_json(file_name: str):
    """Load data from json file."""
    with open(get_file_path(file_name), "r") as file:
        data = json.load(file)
    return data


def _get_abspath(file_name):
    """Get the abspath of the file.

    Raise FileNotFoundError when not found in any subfolder

    """
    for sub_dir in ("examples/configs", "examples/templates"):
        p = os.path.join(_package_path(sub_dir), file_name)
        if glob(formatted_to_asterisked(p)):
            return p
    raise FileNotFoundError(f"Cannot find {file_name}")


def _package_path(sub_directory):
    """Get the abs path of the requested sub folder."""
    return importlib_resources.files("alea") / sub_directory


def formatted_to_asterisked(formatted, wildcards: Optional[Union[str, List[str]]] = None):
    """Convert formatted string to asterisk Sometimes a parameter(usually shape parameter) is not
    specified in formatted string, this function replace the parameter with asterisk.

    Args:
        formatted (str): formatted string
        wildcards (str or list, optional (default=None)):
            wildcards to be replaced with asterisk.

    Returns:
        str: asterisked string

    Examples:
        >>> formatted_to_asterisked("a_{a:.2f}_b_{b:d}")
        "a_*_b_*"
        >>> formatted_to_asterisked("a_{a:.2f}_b_{b:d}", wildcards="a")
        "a_*_b_{b:d}"

    """
    # find all wildcards if wildcards is None
    if wildcards is None:
        return re.sub("{(.*?)}", "*", formatted)

    # convert wildcards to list if wildcards is a string
    if isinstance(wildcards, str):
        wildcards = [wildcards]
    else:
        if not isinstance(wildcards, list):
            raise ValueError(
                f"wildcards must be a string or list of strings, not {type(wildcards)}"
            )

    # replace wildcards with asterisk
    asterisked = formatted
    for found in re.findall("{(.*?)}", formatted):
        for wildcard in wildcards:
            if wildcard in found:
                asterisked = asterisked.replace("{" + found + "}", "*")
    return asterisked


def get_file_path(fname, folder_list: Optional[List[str]] = None):
    """Find the full path to the resource file Try 5 methods in the following order.

    #. fname begin with '/', return absolute path
    #. folder begin with '/', return folder + name
    #. can get file from _get_abspath, return alea internal file path
    #. can be found in local installed ntauxfiles, return ntauxfiles absolute path
    #. can be downloaded from MongoDB, download and return cached path

    Args:
        fname (str): file name
        folder_list (list, optional (default=None)):
            list of possible base folders. Ordered by priority.
            The function will search for file from the first folder in the list,
            and return the first found file immediately without searching the rest folders.

    Returns:
        str: full path to the resource file

    """
    if os.path.exists(fname):
        return fname

    if folder_list is None:
        folder_list = []
    # 1. From absolute path
    # Usually Config.default is a absolute path
    if fname.startswith("/"):
        return fname

    # 2. From local folder
    # Use folder as prefix
    for folder in folder_list:
        if folder.startswith("/"):
            fpath = os.path.join(folder, fname)
            if glob(formatted_to_asterisked(fpath)):
                logging.info(f"Load {fname} successfully from {fpath}")
                return fpath

    # 3. From alea internal files
    try:
        return _get_abspath(fname)
    except FileNotFoundError:
        pass

    # raise error when can not find corresponding file
    raise RuntimeError(f"Can not find {fname}, please check your file system")


def get_template_folder_list(likelihood_config, extra_template_path: Optional[str] = None):
    """Get a list of template_folder from likelihood_config."""
    if "template_folder" not in likelihood_config:
        # return empty list if template_folder is not specified
        likelihood_config["template_folder"] = []
    if isinstance(likelihood_config["template_folder"], str):
        template_folder_list = [likelihood_config["template_folder"]]
    elif isinstance(likelihood_config["template_folder"], list):
        template_folder_list = likelihood_config["template_folder"]
    elif likelihood_config["template_folder"] is None:
        template_folder_list = []
    else:
        raise ValueError("template_folder must be either a string or a list of strings.")
    # Add extra_template_path to the end of template_folder_list
    if extra_template_path is not None:
        template_folder_list.append(extra_template_path)
    return list(set(template_folder_list))


def asymptotic_critical_value(
    confidence_interval_kind: str, confidence_level: float, degree_of_freedom: Optional[int] = None
):
    """Return the critical value for the confidence interval.

    Args:
        confidence_interval_kind (str): confidence interval kind, either 'lower', 'upper' or
            'central'
        confidence_level (float): confidence level
        degree_of_freedom (int, optional (default=None)): degree of freedom

    Returns:
        float: critical value

    Raises:
        ValueError: if confidence_interval_kind is not 'lower', 'upper' or 'central'
        ValueError: if degree_of_freedom is not None and not 1, when confidence_interval_kind is
            'lower' or 'upper'

    """
    if confidence_interval_kind in {"lower", "upper"}:
        if (degree_of_freedom is not None) and (degree_of_freedom != 1):
            raise ValueError(
                f"degree_of_freedom must be 1 for {confidence_interval_kind} confidence interval"
            )
        critical_value = chi2(1).isf(2 * (1.0 - confidence_level))
    elif confidence_interval_kind == "central":
        if degree_of_freedom is None:
            critical_value = chi2(1).isf(1.0 - confidence_level)
        else:
            critical_value = chi2(degree_of_freedom).isf(1.0 - confidence_level)
    else:
        raise ValueError(
            f"confidence_interval_kind must be either 'lower', 'upper' or 'central', "
            f"not {confidence_interval_kind}"
        )
    return critical_value


def within_limits(value, limits):
    """Returns True if value is within limits."""
    if limits is None:
        return True
    elif limits[0] is None:
        return value <= limits[1]
    elif limits[1] is None:
        return value >= limits[0]
    else:
        return limits[0] <= value <= limits[1]


def clip_limits(value) -> Tuple[float, float]:
    """Clip limits to be within [-MAX_FLOAT, MAX_FLOAT] by converting None to -MAX_FLOAT and
    MAX_FLOAT."""
    if value is None:
        value = [-MAX_FLOAT, MAX_FLOAT]
    else:
        if value[0] is None:
            value[0] = -MAX_FLOAT
        if value[1] is None:
            value[1] = MAX_FLOAT
    return value


def can_assign_to_typing(value_type, target_type) -> bool:
    """Check if value_type can be assigned to target_type. This is useful when converting Runner's
    argument into strings.

    Args:
        value_type: type of the value, might be float, int, etc.
        target_type: type of the target, might be Optinal, Union, etc.

    """
    if get_origin(target_type) is Union:
        # If the target type is a Union (like Optional)
        return any(can_assign_to_typing(value_type, t) for t in get_args(target_type))
    else:
        if get_origin(target_type):
            return issubclass(value_type, get_origin(target_type))
        else:
            return issubclass(value_type, target_type)


def add_i_batch(filename: str) -> str:
    """Add i_batch to filename."""
    if "i_batch" in filename:
        raise ValueError("i_batch already in filename")
    fpat_split = os.path.splitext(filename)
    return fpat_split[0] + "_{i_batch:d}" + fpat_split[1]


def search_filename_pattern(filename: str) -> str:
    """Return pattern for a given existing filename. This is needed because sometimes the filename
    is not appended by "_{i_batch:d}". We need to distinguish between the two cases and return the
    correct pattern.

    Returns:
        str: existing pattern for filename, either filename or filename w/ inserted "_*"

    """
    # try to add a * to the filename to read all the files
    fpat_split = os.path.splitext(filename)
    _filename = fpat_split[0] + "_*" + fpat_split[1]
    if len(sorted(glob(_filename))) != 0:
        pattern = _filename
    else:
        pattern = filename
    filename_list = sorted(glob(pattern))
    if len(filename_list) == 0:
        raise ValueError(f"Can not find any output file {filename}!")
    return pattern


def get_metadata(output_filename_pattern: str) -> list:
    """Get metadata from output files."""
    output_filename_list = sorted(glob(output_filename_pattern))
    metadata_list = []
    for _output_filename in output_filename_list:
        with h5py.File(_output_filename, "r", libver="latest", swmr=True) as ipt:
            metadata = dict(
                zip(
                    ipt.attrs.keys(),
                    [json.loads(ipt.attrs[key]) for key in ipt.attrs.keys()],
                )
            )
        metadata_list.append(metadata)
    return metadata_list


def can_expand_grid(variations: dict) -> bool:
    """Check if variations can be expanded into a grid.

    Example:
        >>> can_expand_grid({'a': [1, 2], 'b': [3, 4]})
        True

    """

    # check if all values are lists or no values is list
    is_list = [isinstance(value, list) for value in variations.values()]
    if {True, False}.issubset(is_list):
        raise ValueError(
            "If some values in variations are lists, "
            "all values must be lists or no values is list. "
            f"But you are mixing lists and non-lists in {variations}."
        )
    if all(is_list):
        return True
    else:
        return False


def expand_grid_dict(variations: List[Union[dict, str]]) -> List[Union[dict, str]]:
    """Expand dict into a list of dict, according to the itertools.product method, if necessary.

    Args:
        variations (list): variations to be expanded

    Example:
        >>> expand_grid_dict(["free", {"a": 1, "b": 3}, {"a": 'np.arange(1, 3)', "b": [3, 4]}])
        [
            "free",
            {"a": 1, "b": 3},
            {"a": 1, "b": 3},
            {"a": 1, "b": 4},
            {"a": 2, "b": 3},
            {"a": 2, "b": 4},
        ]

    """

    result = cast(List[Union[dict, str]], [])
    for v in variations:
        # convert str to list first
        if isinstance(v, dict):
            v = evaluate_numpy_scipy_expression_in_dict(v)
        # expand to grid if necessary
        if isinstance(v, dict) and can_expand_grid(v):
            result += convert_to_vary(v)
        else:
            result.append(v)
    return result


def convert_variations(variations: dict, iteration) -> list:
    """Convert variations to a list of dict, according to the iteration method.

    Args:
        variations (dict): variations to be converted
        iteration: iteration method, either zip or itertools.product

    Returns:
        list: a list of dict

    """

    # evaluate numpy and scipy expression in variations
    variations = evaluate_numpy_scipy_expression_in_dict(variations)

    # expand to grid if necessary
    for k, v in variations.items():
        if not isinstance(v, list):
            raise ValueError(f"variations {k} must be a list, not {v} with {type(v)}")
        variations[k] = expand_grid_dict(v)
    result = [dict(zip(variations, deepcopy(t))) for t in iteration(*variations.values())]
    if result:
        return result
    else:
        return [{}]


def convert_to_zip(to_zip: Dict[str, List]) -> List[Dict[str, Any]]:
    """Convert dict into a list of dict, according to the zip method.

    Example:
        >>> convert_to_zip({'a': [1, 2], 'b': [3, 4]})
        [{'a': 1, 'b': 3}, {'a': 2, 'b': 4}]

    """
    return convert_variations(to_zip, zip)


def convert_to_vary(to_vary: Dict[str, List]) -> List[Dict[str, Any]]:
    """Convert dict into a list of dict, according to the itertools.product method.

    Example:
        >>> convert_to_vary({'a': [1, 2], 'b': [3, 4]})
        [{'a': 1, 'b': 3}, {'a': 1, 'b': 4}, {'a': 2, 'b': 3}, {'a': 2, 'b': 4}]

    """
    return convert_variations(to_vary, itertools.product)


def convert_to_in_common(in_common: Dict[str, Any]) -> Dict[str, Any]:
    """Expand the values in in_common, according to the itertools.product method, if necessary. This
    usually happens to the hypotheses.

    Example:
        >>> convert_to_in_common({'hypotheses': ['free', {'a': [1, 2], 'b': [3, 4]}]})
        {
            "hypotheses": [
                "free",
                {"a": 1, "b": 3},
                {"a": 1, "b": 4},
                {"a": 2, "b": 3},
                {"a": 2, "b": 4},
            ]
        }

    """
    for k, v in in_common.items():
        if isinstance(v, list) and (k != "hypotheses") and (k != "confidence_levels"):
            raise ValueError(
                f"except hypotheses, in_common can not contain list, "
                f"you might need to put {(k, v)} in to_zip or to_vary"
            )
    if "hypotheses" in in_common:
        in_common["hypotheses"] = expand_grid_dict(in_common["hypotheses"])
    return in_common


def compute_variations(to_zip, to_vary, in_common) -> list:
    """Compute variations of Runner from to_zip, to_vary and in_common. By priority, the order is
    to_zip, to_vary, in_common. The values in to_zip will overwrite the keys in to_vary and
    in_common. The values in to_vary will overwrite the keys in in_common.

    Args:
        to_zip (dict): variations to be zipped
        to_vary (dict): variations to be varied
        in_common (dict): variations in common

    Returns:
        list: a list of dict

    """
    zipped = convert_to_zip(to_zip=to_zip)
    varied = convert_to_vary(to_vary=to_vary)

    combined = [
        {**convert_to_in_common(in_common), **v, **z} for z, v in itertools.product(zipped, varied)
    ]
    return combined


def make_hashable(obj):
    """Convert a container hierarchy into one that can be hashed.

    See http://stackoverflow.com/questions/985294

    """
    if isinstance(obj, Mapping):
        # Convert immutabledict etc for json decoding
        obj = dict(obj)
    try:
        hash(obj)
    except TypeError:
        if isinstance(obj, dict):
            return tuple((k, make_hashable(v)) for (k, v) in sorted(obj.items()))
        elif isinstance(obj, np.ndarray):
            return tuple(obj.tolist())
        elif hasattr(obj, "__iter__"):
            return tuple(make_hashable(o) for o in obj)
        else:
            raise TypeError("Can't make_hashable object of type %r" % type(obj))
    else:
        return obj


def deterministic_hash(thing, length=10):
    """Return a base32 lowercase string of length determined from hashing a container hierarchy.

    Edited from strax: strax/utils.py

    """
    hashable = make_hashable(thing)
    jsonned = json.dumps(hashable)
    # disable bandit
    digest = sha256(jsonned.encode("ascii")).digest()
    return b32encode(digest)[:length].decode("ascii").lower()


def signal_multiplier_estimator(
    signal: np.ndarray,
    background: np.ndarray,
    data: np.ndarray,
    iteration=100,
    diagnostic=False,
) -> float:
    """Estimate the best-fit signal multiplier using perturbation theory. The method tries to solve
    the critial point of the likelihood function by perturbation theory, where the likelihood
    function is defined as the binned Poisson likelihood function, given signal, background models
    and data.

    Args:
        signal (np.ndarray): signal model
        background (np.ndarray): background model
        data (np.ndarray): data array
        iteration (int, optional (default=100)): number of iterations
    Returns:
        float: best-fit signal multiplier

    """
    mask = (signal > 0) | (background > 0)
    if np.any(data[~mask] > 0):
        raise ValueError("Data has non-zero values where signal and background is zero.")

    sig = signal[mask].ravel()
    bkg = background[mask].ravel()
    obs = data[mask].ravel()

    @np.errstate(invalid="ignore", divide="ignore")
    def correction_on_multiplier(x):
        exp = sig * x + bkg
        return np.sum(np.where(exp > 0, (obs / exp - 1) * sig, 0)) / np.sum(
            np.where(exp > 0, obs * sig**2 / exp**2, 0)
        )

    # For underfluctutation case, the best-fit multiplier could be negative
    # in which case the perturbation theory may not converge or be negative.
    # Thus we clip it to be non-negative.
    x = np.sum(obs - bkg) / np.sum(sig)
    xs = [x]
    for _ in range(iteration):
        x += correction_on_multiplier(x)
        x = np.clip(x, 0, None)
        xs.append(x)
    if diagnostic:
        plt.plot(xs, marker=".")
        plt.xlabel("Iteration")
        plt.ylabel("x")
    return x


class IndexMorpher(Morpher):
    """IndexMorpher is a morpher which applies no interpolation."""

    def get_anchor_points(self, bounds, n_models=None):
        grid = [par.keys() for _, (par, _, _) in self.shape_parameters.items()]
        return list(product(*grid))

    def make_interpolator(self, f, extra_dims, anchor_models):
        return lambda z: f(anchor_models[tuple(z)])


blueice.pdf_morphers.MORPHERS["IndexMorpher"] = IndexMorpher



from copy import deepcopy
from unittest import TestCase

from alea.utils import load_yaml
from alea.parameters import Parameters


class TestParameters(TestCase):
    """Test of the Parameters class."""

    def setUp(self):
        """Initialise the Parameters instance."""
        filenames = [
            "unbinned_wimp_statistical_model.yaml",
            "unbinned_wimp_statistical_model_mass_dependent_efficiency.yaml",
        ]
        configs = []
        for fn in filenames:
            configs.append(load_yaml(fn)["parameter_definition"])
        self.configs = configs

        parameters_list = []
        for config in self.configs:
            parameters_list.append(Parameters.from_config(config))
        self.parameters_list = parameters_list

    def test_from_list(self):
        """Test of the from_list method."""
        for config, parameters in zip(self.configs, self.parameters_list):
            only_name_parameters = Parameters.from_list(config.keys())
            # it is false because only names are assigned
            self.assertFalse(only_name_parameters == parameters)

    def test___repr__(self):
        """Test of the __repr__ method."""
        for parameters in self.parameters_list:
            for p in parameters:
                if not isinstance(repr(p), str):
                    raise ValueError("The __repr__ method does not return the correct string.")
            if not isinstance(repr(parameters), str):
                raise TypeError("The __repr__ method does not return a string.")

    def test_deep_copyable(self):
        """Test of whether Parameters instance can be deepcopied."""
        for parameters in self.parameters_list:
            if deepcopy(parameters) != parameters:
                raise ValueError("Parameters instance cannot be correctly deepcopied.")

    def test_conditional_parameter(self):
        """Test of the ConditionalParameter class."""
        config = self.configs[1]
        parameters = self.parameters_list[1]
        nominal_wimp_mass = config["wimp_mass"]["nominal_value"]
        signal_eff_uncert_dict = config["signal_efficiency"]["uncertainty"]

        # Directly accessing the property should return the value
        # under nominal conditions
        val = parameters.signal_efficiency.uncertainty
        expected_val = signal_eff_uncert_dict[nominal_wimp_mass]
        self.assertEqual(val, expected_val)

        # Calling without kwargs should return the value
        # under nominal conditions
        val = parameters.signal_efficiency().uncertainty
        expected_val = signal_eff_uncert_dict[nominal_wimp_mass]
        self.assertEqual(val, expected_val)

        # Calling with kwargs should return the value under
        # the specified conditions
        for wimp_mass, expected_val in signal_eff_uncert_dict.items():
            val = parameters.signal_efficiency(wimp_mass=wimp_mass).uncertainty
            self.assertEqual(val, expected_val)



import inspect
import warnings
from pydoc import locate
from copy import deepcopy
from typing import List, Tuple, Callable, Optional, Union
from itertools import product

import numpy as np
from scipy.optimize import brentq
from iminuit import Minuit
from blueice.likelihood import _needs_data
from inference_interface import toydata_to_file

from alea.parameters import Parameters
from alea.utils import within_limits, clip_limits, asymptotic_critical_value, ReadOnlyDict

_DEFAULT_FIT_STRATEGY = {
    "disable_index_fitting": False,
    "max_index_fitting_iter": 10,
    "minimizer_routine": "migrad",
    "minuit_strategy": 1,
    "refit_invalid": True,
}


class StatisticalModel:
    """Class that defines a statistical model.

    - The statisical model contains two parts that you must define yourself:
        - a likelihood function
            ll(self, parameter_1, parameter_2... parameter_n):
            A function of a set of named parameters which
            return a float expressing the loglikelihood for observed data given these parameters.
        - a data generation function
            generate_data(self, parameter_1, parameter_2... parameter_n):
            A function of the same set of named parameters return a full data set.
    - Methods that you must implement:
        - _ll
        - _generate_data
    - Methods that you may implement:
        - get_expectation_values
    - Methods that already exist here:
        - ll
        - store_data
        - fit
        - get_parameter_list
        - confidence_interval

    The public methods generate_data and ll, as the names suggested,
    depend on private methods _generate_data, and _ll respectively.

    Attributes:
        data: data of the model
        _data: data of the model
        _confidence_level: confidence level for confidence intervals
        _confidence_interval_kind: kind of confidence interval to compute
        parameters: parameters of the model
        confidence_interval_threshold: threshold for confidence interval
        is_data_set (bool): True if data is set

    Args:
        parameter_definition (dict or list, optional (default=None)):
            definition of the parameters of the model
        confidence_level (float, optional (default=0.9)):
            confidence level for confidence intervals
        confidence_interval_kind (str, optional (default="central")):
            kind of confidence interval to compute
        confidence_interval_threshold (Callable[[float], float], optional (default=None)):
            threshold for confidence interval
        data (dict or list, optional (default=None)): pre-set data of the model
        fit_strategy (dict, optional (default=None)): strategy for the fit,
            see _DEFAULT_FIT_STRATEGY for possible settings

    Raise:
        RuntimeError: if you try to instantiate the StatisticalModel class directly
        NotImplementedError: if you do not implement the likelihood function or the data generation

    """

    def __init__(
        self,
        parameter_definition: Optional[Union[dict, list]] = None,
        confidence_level: float = 0.9,
        confidence_interval_kind: str = "central",  # one of central, upper, lower
        confidence_interval_threshold: Optional[Callable[[float], float]] = None,
        asymptotic_dof: Optional[int] = 1,
        data: Optional[Union[dict, list]] = None,
        fit_strategy: Optional[dict] = None,
        **kwargs,
    ):
        """Initialize a statistical model."""
        if type(self) == StatisticalModel:  # noqa: E721
            raise RuntimeError(
                "You cannot instantiate the StatisticalModel class directly, "
                "you must use a subclass where the likelihood function and data generation "
                "method are implemented"
            )

        # following https://github.com/JelleAalbers/blueice/blob/
        # 7c10222a13227e78dc7224b1a7e56ff91e4a8043/blueice/likelihood.py#L97
        self.is_data_set = False
        if data is not None:
            self.data = data
        self._confidence_level = confidence_level
        if confidence_interval_kind not in {"central", "upper", "lower"}:
            raise ValueError("confidence_interval_kind must be one of central, upper, lower")
        self._confidence_interval_kind = confidence_interval_kind
        self.confidence_interval_threshold = confidence_interval_threshold
        self.asymptotic_dof = asymptotic_dof
        self._fit_strategy = _DEFAULT_FIT_STRATEGY
        if fit_strategy is not None:
            self._fit_strategy.update(fit_strategy)
        nominal_values = kwargs.get("nominal_values", None)
        self._define_parameters(parameter_definition, nominal_values)

        self._check_ll_and_generate_data_signature()

    def _define_parameters(self, parameter_definition, nominal_values=None):
        """Initialize the parameters of the model."""
        if parameter_definition is None:
            self.parameters = Parameters()
        elif isinstance(parameter_definition, dict):
            # if nominal_values are given, overwrite the ones in parameter_definition
            if nominal_values is not None:
                for name, definition in parameter_definition.items():
                    if name in nominal_values:
                        definition["nominal_value"] = nominal_values[name]
            self.parameters = Parameters.from_config(parameter_definition)
        elif isinstance(parameter_definition, list):
            self.parameters = Parameters.from_list(parameter_definition)
            if nominal_values is not None:
                self.parameters.set_nominal_values(**nominal_values)
        else:
            raise RuntimeError("parameter_definition must be dict or list")

    def _check_ll_and_generate_data_signature(self):
        """Check that the likelihood and generate_data functions have the same signature."""
        ll_params = set(inspect.signature(self._ll).parameters)
        generate_data_params = set(inspect.signature(self._generate_data).parameters)
        if ll_params != generate_data_params:
            raise AssertionError("ll and generate_data must have the same signature (parameters)")

    def _ll(self, **kwargs) -> float:
        """Likelihood function, return the loglikelihood for the given parameters."""
        raise NotImplementedError(
            "You must write a likelihood function (_ll) for your statistical model"
            " or use a subclass where it is written for you"
        )

    def _generate_data(self, **kwargs):
        """Generate data for the given parameters."""
        raise NotImplementedError(
            "You must write a data-generation method (_generate_data) for your statistical model"
            " or use a subclass where it is written for you"
        )

    @_needs_data
    def ll(self, **kwargs) -> float:
        """Likelihod function, returns the loglikelihood for the given parameters. The parameters
        are passed as keyword arguments, positional arguments are not possible. If a parameter is
        not given, the default value is used.

        Keyword Args:
            kwargs: keyword arguments for the parameters

        Returns:
            float: likelihood value

        """
        parameters = self.parameters(**kwargs)
        return self._ll(**parameters)

    def generate_data(self, **kwargs) -> Union[dict, list]:
        """Generate data for the given parameters. The parameters are passed as keyword arguments,
        positional arguments are not possible. If a parameter is not given, the default value is
        used.

        Raises:
            ValueError: If the parameters are not within the fit limits

        Returns:
            dict or list: generated data

        Caution:
            This implementation won't allow you to call generate_data by positional arguments.

        """
        if not self.parameters.values_in_fit_limits(**kwargs):
            raise ValueError("Values are not within fit limits")
        generate_values = self.parameters(**kwargs)
        return self._generate_data(**generate_values)

    @property
    def data(self):
        """Simple getter for a data-set-- mainly here so it can be over-ridden for special needs.

        Data-sets are expected to be in the form of a list of one or more structured arrays,
        representing the data-sets of one or more likelihood terms.

        """
        if self._data is None:
            raise RuntimeError("data has not been assigned this statistical model!")
        return self._data

    @data.setter
    def data(self, data):
        """Data setter."""
        self._data = data
        self.is_data_set = True

    def store_data(
        self,
        file_name,
        data_list,
        data_name_list: Optional[List[str]] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Store a list of datasets.
        (each on the form of a list of one or more structured arrays or dicts)
        Using inference_interface, but included here to allow over-writing.
        The structure would be: ``[[datasets1], [datasets2], ..., [datasetsn]]``,
        where each of datasets is a list of structured arrays.
        If you specify, it is set, if not it will read from ``self.get_likelihood_term_names``.
        If not defined, it will be ``["0", "1", ..., "n-1"]``. The metadata is optional.

        Args:
            file_name (str): name of the file to store the data in
            data_list (list): list of datasets
            data_name_list (list, optional (default=None)): list of names of the datasets.
                If None, it will be read from self.get_likelihood_term_names
            metadata (dict, optional (default=None)): metadata to store with the data.
                If None, no metadata is stored.
        """
        if all([isinstance(d, dict) for d in data_list]) or all(
            [isinstance(d, ReadOnlyDict) for d in data_list]
        ):
            _data_list = [list(d.values()) for d in data_list]
        elif all([isinstance(d, list) for d in data_list]):
            _data_list = data_list
        else:
            raise ValueError(
                "Unsupported mixed toydata format! "
                "toydata should be a list of dict or a list of list",
            )

        if data_name_list is None:
            if hasattr(self, "likelihood_names"):
                data_name_list = self.likelihood_names
            else:
                data_name_list = ["{:d}".format(i) for i in range(len(_data_list[0]))]

        kw = {"metadata": metadata} if metadata is not None else dict()
        if len(_data_list[0]) != len(data_name_list):
            raise ValueError("The number of data sets and data names must be the same")
        toydata_to_file(file_name, _data_list, data_name_list, **kw)

    def set_fit_guesses(self, **fit_guesses):
        """Set the fit guesses for parameters.

        Keyword Args:
            fit_guesses (dict): A dict of parameter names and values.

        """
        self.parameters.set_fit_guesses(**fit_guesses)

    def get_expectation_values(self, **parameter_values):
        """Get the expectation values of the measurement.

        Args:
            parameter_values: values of the parameters

        """
        raise NotImplementedError("get_expectation_values is optional to implement")

    @property
    def nominal_expectation_values(self):
        """Nominal expectation values for the sources of the likelihood.

        For this to work, you must implement ``get_expectation_values``.

        """
        return self.get_expectation_values()  # no kwargs for nominal

    def get_likelihood_term_from_name(self, likelihood_name: str) -> int:
        """Return the index of a likelihood term if the likelihood has several names.

        Args:
            likelihood_name (str): name of the likelihood term

        Returns:
            int: index of the likelihood term

        """
        if hasattr(self, "likelihood_names"):
            likelihood_names = self.likelihood_names
            return {n: i for i, n in enumerate(likelihood_names)}[likelihood_name]
        else:
            raise NotImplementedError("The attribute likelihood_names is not defined.")

    def get_parameter_list(self):
        """Return a set of all parameters that the generate_data and likelihood accepts."""
        return self.parameters.names

    def make_objective(self):
        """Make a function that can be passed to Minuit.

        Returns:
            Callable: function that can be passed to Minuit

        """

        def cost(args):
            # Get the arguments from args
            call_kwargs = {}
            for i, k in enumerate(self.parameters.names):
                call_kwargs[k] = args[i]
            # for optimization, we want to minimize the negative log-likelihood
            return self.ll(**call_kwargs) * -1

        return cost

    @_needs_data
    def fit(
        self, verbose: Optional[bool] = False, fit_strategy: Optional[dict] = None, **kwargs
    ) -> Tuple[dict, float]:
        """Fit the model to the data by maximizing the likelihood. Return a dict containing best-fit
        values of each parameter, and the value of the likelihood evaluated there. While the
        optimization is a minimization, the likelihood returned is the __maximum__ of the
        likelihood.

        Args:
            verbose (bool): if True, print the Minuit object
            fit_strategy (dict): override the default fit strategy defined
                in the model (model.fit_strategy). Possible settings are:
                - minimizer_routine (str): the minimizer routine to use, either
                    "migrad", "simplex", or "simplex_migrad" (first run simplex, then migrad).
                - minuit_strategy (int): strategy for Minuit, can be 0, 1, or 2. The higher the
                    number, the more precise the fit but also the slower.
                - refit_invalid (bool): if True, refit with the simplex_migrad routine
                    and strategy 2 if the optimization does not converge the first time.
                - disable_index_fitting (bool): if True, disable the index fitting
                    even if the model has index parameters.
                - max_index_fitting_iter (int): maximum number of iterations for index fitting

        Returns:
            dict, float: best-fit values of each parameter,
            and the value of the likelihood evaluated there

        """
        fit_strategy = self._get_fit_strategy(fit_strategy)
        fixed_parameters = list(kwargs.keys())
        guesses = self.parameters.fit_guesses
        guesses.update(kwargs)
        if not self.parameters.values_in_fit_limits(**guesses):
            raise ValueError("Initial guesses are not within fit limits")
        defaults = self.parameters(**guesses)

        cost = self.make_objective()

        # Make the Minuit object
        m = Minuit(MinuitWrap(cost, parameters=self.parameters), **defaults)
        m.errordef = Minuit.LIKELIHOOD
        m.strategy = fit_strategy["minuit_strategy"]
        fixed_params = [] if fixed_parameters is None else fixed_parameters
        fixed_params += self.parameters.not_fittable
        for par in fixed_params:
            m.fixed[par] = True

        # Get the index parameters, which could have problem if simply using migrad
        index_parameters = [
            p for p in self.parameters if p.ptype == "index" and p.name not in fixed_params
        ]
        if fit_strategy["disable_index_fitting"] or (len(index_parameters) == 0):
            m = self._standard_fit(m, fit_strategy["minimizer_routine"])
            if not m.valid and fit_strategy["refit_invalid"]:
                # try to refit with more precision
                m.strategy = 2
                m = self._standard_fit(m, "simplex_migrad")
        else:
            m = self._index_mixing_fit(
                m,
                index_parameters,
                fit_strategy["max_index_fitting_iter"],
                verbose,
                fit_strategy["minimizer_routine"],
            )

        self.minuit_object = m
        if verbose:
            print(m)
        # alert! This gives the _maximum_ likelihood
        return m.values.to_dict(), -1 * m.fval

    def _get_fit_strategy(self, fit_strategy) -> dict:
        # override the default fit strategy
        if fit_strategy is None:
            fit_strategy = self._fit_strategy
        else:
            # check if keys are valid
            for key in fit_strategy.keys():
                if key not in _DEFAULT_FIT_STRATEGY:
                    raise ValueError(f"Unknown key {key} in fit_strategy")
            # fill the gaps of fit_strategy with self.fit_strategy
            for key, value in self._fit_strategy.items():
                fit_strategy.setdefault(key, value)
        return fit_strategy

    def _standard_fit(self, m, minimizer_routine):
        m = self._run_minimizer_routine(m, minimizer_routine)
        return m

    def _index_mixing_fit(
        self, m, index_parameters, max_index_fitting_iter, verbose, minimizer_routine
    ):
        index_anchors = [p.blueice_anchors for p in index_parameters]
        index_names = [p.name for p in index_parameters]
        index_grid = [
            {index_names[i]: anchor[i] for i in range(len(anchor))}
            for anchor in product(*index_anchors)
        ]

        # We fix the index parameters in migrad
        for par in index_names:
            m.fixed[par] = True

        # We firstly do optimization on other parameters with index parameters
        # fixed to their initial guesses. Then we grid search over the index
        # parameters given the optimized parameters. We repeat the optimization
        # and grid search until the optimization converges
        for itr in range(max_index_fitting_iter):
            m = self._run_minimizer_routine(m, minimizer_routine)
            # Find the best-fit index parameters
            lls = np.zeros(len(index_grid))
            for i in range(len(lls)):
                best_fit_params = m.values.to_dict()
                best_fit_params.update(index_grid[i])
                lls[i] = self.ll(**best_fit_params)
            for var in index_names:
                m.values[var] = index_grid[np.argmax(lls)][var]

            # Calculating Hessian will update the validity of
            # the fitting given the new index parameters
            m.hesse()
            if m.valid:
                break

        if verbose and itr == max_index_fitting_iter - 1:
            print(
                "The index searching iteration times reached the maximum! "
                "The optimization could not converge!"
            )
        return m

    @staticmethod
    def _run_minimizer_routine(m, minimizer_routine):
        if minimizer_routine == "migrad":
            m.migrad()
        elif minimizer_routine == "simplex":
            m.simplex()
        elif minimizer_routine == "simplex_migrad":
            m.simplex().migrad()
        else:
            raise ValueError(
                "minimizer_routine must be one of "
                "'migrad', 'simplex', 'simplex_migrad' "
                f"but got {minimizer_routine}"
            )
        return m

    def _confidence_interval_checks(
        self,
        poi_name: str,
        parameter_interval_bounds: Optional[Tuple[float, float]] = None,
        confidence_level: Optional[float] = None,
        confidence_interval_kind: Optional[str] = None,
        confidence_interval_threshold: Optional[Callable[[float], float]] = None,
        asymptotic_dof: Optional[int] = None,
        **kwargs,
    ) -> Tuple[str, Callable[[float], float], Tuple[float, float]]:
        """Helper function for confidence_interval that does the input checks and return bounds.

        Args:
            poi_name (str): name of the parameter of interest
            parameter_interval_bounds (Tuple[float, float], optional (default=None)):
                range in which to search for the confidence interval edges
            confidence_level (float, optional (default=None)):
                confidence level for confidence intervals
            confidence_interval_kind (str, optional (default=None)):
                kind of confidence interval to compute

        Returns:
            Tuple[str, Callable[[float], float], Tuple[float, float]]:
                confidence interval kind, confidence interval threshold, parameter interval bounds

        """
        if confidence_level is None:
            confidence_level = self._confidence_level
        if confidence_interval_kind is None:
            confidence_interval_kind = self._confidence_interval_kind

        if (confidence_level < 0) or (confidence_level > 1):
            raise ValueError("confidence_level must be between 0 and 1")

        parameter_of_interest = self.parameters[poi_name]
        if not parameter_of_interest.fittable:
            raise ValueError("The parameter of interest must be fittable")
        if poi_name in kwargs:
            raise ValueError("You cannot set the parameter you're constraining")

        if parameter_interval_bounds is None:
            parameter_interval_bounds = parameter_of_interest.parameter_interval_bounds
        else:
            value = parameter_interval_bounds
            parameter_of_interest._check_parameter_interval_bounds(value)
            parameter_interval_bounds = clip_limits(value)

        # define threshold if none is defined:
        if confidence_interval_threshold is None:
            if self.confidence_interval_threshold is not None:
                confidence_interval_threshold = self.confidence_interval_threshold
            else:
                # use asymptotic thresholds assuming the test statistic is Chi2 distributed
                if asymptotic_dof is None:
                    if self.asymptotic_dof is not None:
                        degree_of_freedom = self.asymptotic_dof
                    else:
                        degree_of_freedom = 1
                else:
                    if self.asymptotic_dof is not None:
                        if asymptotic_dof != self.asymptotic_dof:
                            warnings.warn(
                                f"You set asymptotic_dof as {asymptotic_dof}, "
                                f"which is different from the default value "
                                f"{self.asymptotic_dof}. Be careful!"
                            )
                    degree_of_freedom = asymptotic_dof
                critical_value = asymptotic_critical_value(
                    confidence_interval_kind, confidence_level, degree_of_freedom
                )

                def confidence_interval_threshold(_):
                    return critical_value

        else:
            if self.confidence_interval_threshold is not None:
                raise ValueError(
                    "You cannot set confidence_interval_threshold twice, "
                    "once in the constructor and once in the method call"
                )
        if not callable(confidence_interval_threshold):
            raise ValueError("confidence_interval_threshold must be a callable")

        return confidence_interval_kind, confidence_interval_threshold, parameter_interval_bounds

    def confidence_interval(
        self,
        poi_name: str,
        parameter_interval_bounds: Optional[Tuple[float, float]] = None,
        confidence_level: Optional[float] = None,
        confidence_interval_kind: Optional[str] = None,
        confidence_interval_threshold: Optional[Callable[[float], float]] = None,
        confidence_interval_args: Optional[dict] = None,
        best_fit_args: Optional[dict] = None,
        asymptotic_dof: Optional[int] = None,
        fit_strategy: Optional[dict] = None,
    ) -> Tuple[float, float]:
        """Uses self.fit to compute confidence intervals for a certain named parameter. If the
        parameter is a rate parameter, and the model has expectation values implemented, the bounds
        will be interpreted as bounds on the expectation value, so that the range in the fit is
        parameter_interval_bounds/mus. Otherwise the bound is taken as-is.

        Args:
            poi_name (str): name of the parameter of interest
            parameter_interval_bounds (Tuple[float, float], optional (default=None)): range
                in which to search for the confidence interval edges. May be specified as:
                    - setting the property "parameter_interval_bounds" for the parameter
                    - passing a list here
                    - passing None here, the property of the parameter is used
            confidence_level (float, optional (default=None)):
                confidence level for confidence intervals.
                If None, the default confidence level of the model is used.
            confidence_interval_kind (str, optional (default=None)):
                kind of confidence interval to compute.
                If None, the default kind of the model is used.
            confidence_interval_args (dict, optional (default=None)): Parameters that will be fixed
                in the profile likelihood computation. If None, all fittable parameters
                will be profiled except the poi.
            best_fit_args (dict, optional (default=None)): If you require the "global" best-fit
                used to normalise the profile likelihood ratio to fix fewer parameters than the
                profile likelihood-- mainly used for 1-D slices of higher-dimensional confidence
                volumes, where the global best-fit may not be along the profile.
                If None, will be set to confidence_interval_args.
            asymptotic_dof (int, optional (default=None)): Degrees of freedom for asymptotic
            fit_strategy (dict, optional (default=None)): strategy for the fit,
                see _DEFAULT_FIT_STRATEGY for possible settings.

        """
        if confidence_interval_args is None:
            confidence_interval_args = {}
        if best_fit_args is None:
            best_fit_args = confidence_interval_args
        ci_objects = self._confidence_interval_checks(
            poi_name,
            parameter_interval_bounds,
            confidence_level,
            confidence_interval_kind,
            confidence_interval_threshold,
            asymptotic_dof,
            **confidence_interval_args,
        )
        (
            confidence_interval_kind,
            confidence_interval_threshold,
            parameter_interval_bounds,
        ) = ci_objects

        # best_fit_args only provides the best-fit likelihood
        _, best_ll = self.fit(**best_fit_args, fit_strategy=fit_strategy)
        # the optimization of profile-likelihood under
        # confidence_interval_args provides the best_parameter
        best_result, _ = self.fit(**confidence_interval_args, fit_strategy=fit_strategy)
        best_parameter = best_result[poi_name]
        mask = within_limits(best_parameter, parameter_interval_bounds)
        if not mask:
            raise ValueError(
                f"The best-fit {best_parameter} is outside your confidence interval "
                f"search limits in parameter_interval_bounds {parameter_interval_bounds}."
            )

        # define intersection between likelihood ratio curve and the critical curve:
        def t(hypothesis_value):
            # define the intersection
            # between the profile-log-likelihood curve and the rejection threshold
            _confidence_interval_args = deepcopy(confidence_interval_args)
            _confidence_interval_args[poi_name] = hypothesis_value
            _, ll = self.fit(
                **_confidence_interval_args, fit_strategy=fit_strategy
            )  # ll is + log-likelihood here
            ret = 2.0 * (best_ll - ll)  # likelihood curve "right way up" (smiling)
            # if positive, hypothesis is excluded
            return ret - confidence_interval_threshold(hypothesis_value)

        t_best_parameter = t(best_parameter)

        if t_best_parameter > 0:
            warnings.warn(
                f"CL calculation failed, given fixed parameters {confidence_interval_args}."
            )

        if confidence_interval_kind in {"upper", "central"} and t_best_parameter < 0:
            if t(parameter_interval_bounds[1]) > 0:
                ul = brentq(t, best_parameter, parameter_interval_bounds[1])
            else:
                ul = np.inf
        else:
            ul = np.nan

        if confidence_interval_kind in {"lower", "central"} and t_best_parameter < 0:
            if t(parameter_interval_bounds[0]) > 0:
                dl = brentq(t, parameter_interval_bounds[0], best_parameter)
            else:
                dl = -1 * np.inf
        else:
            dl = np.nan

        return dl, ul

    @staticmethod
    def get_model_from_name(statistical_model: str):
        """Get the statistical model class from a string."""
        statistical_model_class = locate(statistical_model)
        if statistical_model_class is None:
            raise ValueError(f"Could not find {statistical_model}!")
        if not inspect.isclass(statistical_model_class):
            raise ValueError(f"{statistical_model_class} is not a class!")
        if not issubclass(statistical_model_class, StatisticalModel):
            raise ValueError(f"{statistical_model_class} is not a subclass of StatisticalModel!")
        return statistical_model_class


class MinuitWrap:
    """Wrapper for functions to be called by Minuit. Initialized with a function f and a Parameters
    instance.

    Attributes:
        func: function wrapped
        s_args (list): parameter names of the model
        _parameters (dict): parameters and limits of the model

    Args:
        f (Callable): function to be wrapped
        parameters (Parameters): parameters of the model

    """

    def __init__(self, f: Callable, parameters: Parameters):
        """Initialize the wrapper."""
        self.func = f
        self.s_args = parameters.names
        self._parameters = {p.name: p.fit_limits for p in parameters}

    def __call__(self, *args):
        return self.func(args)



from os import remove
from unittest import TestCase

import scipy.stats as sps
import inference_interface

from alea.examples import GaussianModel


gaussian_model_parameter_definition = {
    "mu": {
        "fit_guess": 0.0,
        "fittable": True,
        "nominal_value": 0.0,
        "parameter_interval_bounds": [
            -10,
            10,
        ],
    },
    "sigma": {
        "fittable": False,
        "nominal_value": 1.0,
    },
}


class TestGaussianModel(TestCase):
    """Test of the GaussianModel class."""

    @classmethod
    def setUp(cls):
        """Initialise the GaussianModel instance."""
        cls.model = GaussianModel(parameter_definition=gaussian_model_parameter_definition)

    def test_data_generation(self):
        """Test generation of data."""
        self.model.data = self.model.generate_data(mu=0)
        self.model.generate_data(mu=0, sigma=2)

    def test_data_storage(self):
        """Test storage of data to file and retrieval of data from file."""
        toydata_filename = "simple_data.ii.h5"
        self.model.data = self.model.generate_data(mu=0)
        self.model.store_data(toydata_filename, [self.model.data])
        stored_data = inference_interface.toydata_from_file(toydata_filename)
        assert self.model.data == stored_data[0], "Stored data disagrees with data!"
        remove(toydata_filename)

    def test_fit_result(self):
        """Test fitting of data."""
        self.model.data = self.model.generate_data(mu=0)
        hat_meas = self.model.data[0]["hat_mu"].item()
        best_fit, lf = self.model.fit(sigma=2)
        hat_fit = best_fit["mu"]
        self.assertAlmostEqual(hat_meas, hat_fit)
        self.assertAlmostEqual(lf, sps.norm(hat_fit, 2).logpdf(hat_meas))



from unittest import TestCase

from alea import StatisticalModel


class TestStatisticalModel(TestCase):
    """Test of the GaussianModel class."""

    def test_statistical_model(self):
        with self.assertRaises(
            RuntimeError, msg="Should raise error when directly instantiating StatisticalModel"
        ):
            StatisticalModel()



from os import remove
import pytest
from unittest import TestCase
from copy import deepcopy
import numpy as np

from inference_interface import toydata_from_file

from blueice.likelihood import LogLikelihoodSum
from alea.utils import load_yaml
from alea.models import BlueiceExtendedModel, CustomAncillaryLikelihood


@pytest.mark.usefixtures("rm_cache")
class TestBlueiceExtendedModel(TestCase):
    """Test of the BlueiceExtendedModel class."""

    @classmethod
    def setUp(cls):
        """Initialise the BlueiceExtendedModel instance."""
        cls.configs = [
            load_yaml("unbinned_wimp_statistical_model.yaml"),
            load_yaml("unbinned_wimp_statistical_model_simple.yaml"),
            load_yaml("unbinned_wimp_statistical_model_index_fitting.yaml"),
        ]
        ns = [len(c["likelihood_config"]["likelihood_terms"]) for c in cls.configs]
        cls.n_likelihood_terms = ns
        cls.toydata_filename = "simple_data.ii.h5"
        cls.set_new_models(cls)

    def set_new_models(self):
        """Set a new BlueiceExtendedModel instance."""
        models = []
        for config in self.configs:
            models.append(
                BlueiceExtendedModel(
                    parameter_definition=config["parameter_definition"],
                    likelihood_config=config["likelihood_config"],
                )
            )
        self.models = models

    def test_deep_copyable(self):
        """Test of whether BlueiceExtendedModel instance can be deepcopied."""
        for model in self.models:
            try:
                model.data = model.generate_data()
                deepcopy(model)
            except Exception:
                raise ValueError("BlueiceExtendedModel instance cannot be correctly deepcopied.")

    def get_expectation_values(self):
        # normalization of templates
        nominal_values = {"er": 200, "wimp": 10}
        expectation_values = []
        for config, model in zip(self.configs, self.models):
            this_expectation_dict = {}
            ll_c = config["likelihood_config"]
            for ll_term in ll_c["likelihood_terms"]:
                livetime_parameter = ll_term["livetime_parameter"]
                livetime = model.parameters[livetime_parameter].nominal_value
                for source in ll_term["sources"]:
                    name = source["name"]
                    this_expectation = nominal_values[name]
                    this_expectation *= source.get("histogram_scale_factor", 1.0)
                    this_expectation *= model.parameters[f"{name}_rate_multiplier"].nominal_value
                    this_expectation *= livetime
                    if name in this_expectation_dict:
                        this_expectation_dict[name] += this_expectation
                    else:
                        this_expectation_dict[name] = this_expectation
            expectation_values.append(this_expectation_dict)
        return expectation_values

    def test_all_source_names(self):
        """Test of the all_source_names method."""
        for config, model in zip(self.configs, self.models):
            _source_names = set()
            for ll_t in config["likelihood_config"]["likelihood_terms"]:
                _source_names.update([s["name"] for s in ll_t["sources"]])
            source_names = model.all_source_names
            self.assertEqual(source_names, sorted(_source_names))

    def test_expectation_values(self):
        """Test of the expectation_values method."""

        self.set_new_models()
        naive_expectation_values = self.get_expectation_values()
        for model, naive_vals in zip(self.models, naive_expectation_values):
            expectation_values = model.get_expectation_values()

            # should avoid accidentally set data
            is_data_set = False
            for ll_term in model.likelihood_list[:-1]:
                is_data_set |= ll_term.is_data_set
            if is_data_set:
                raise ValueError("Data should not be set after get_expectation_values.")

            # Check whether the expectation values are correct
            for k, v in expectation_values.items():
                self.assertEqual(v, naive_vals[k])

            # Check whether scaling works
            scaling_factor = 2.0
            new_expectation_values = model.get_expectation_values(
                wimp_rate_multiplier=scaling_factor, er_rate_multiplier=scaling_factor
            )
            for k, v in expectation_values.items():
                self.assertEqual(v * scaling_factor, new_expectation_values[k])

    def test_get_expectation_values_per_likelihood_term(self):
        """Test of the get_expectation_values method with per_likelihood=True."""
        self.set_new_models()
        for model in self.models:
            vals_per = model.get_expectation_values(per_likelihood_term=True)
            vals_total = model.get_expectation_values(per_likelihood_term=False)

            # Manually sum up the per_likelihood_term expectation values
            summed_vals = {}
            for term in vals_per.values():
                for key, value in term.items():
                    summed_vals[key] = summed_vals.get(key, 0) + value

            # Check whether the summed values are equal to the total values
            for key, summed_val in summed_vals.items():
                self.assertEqual(summed_val, vals_total[key])

    def test_store_data(self):
        """Test of the generate_data method."""
        for model, n in zip(self.models, self.n_likelihood_terms):
            data = model.generate_data()
            model.data = data
            data_list_of_dict = [data]
            data_list_of_ordereddict = [model.data]
            data_list_of_list = [[data[k] for k in data.keys()]]
            for d in [data_list_of_dict, data_list_of_ordereddict, data_list_of_list]:
                model.store_data(self.toydata_filename, d)
                remove(self.toydata_filename)

    def test_generate_data(self):
        """Test of the generate_data method."""
        for model, n in zip(self.models, self.n_likelihood_terms):
            data = model.generate_data()
            self.assertEqual(len(data), n + 2)
            if not (("ancillary" in data) and ("generate_values" in data)):
                raise ValueError("Data does not contain ancillary and generate_values.")
            for k, v in data.items():
                if k in {"ancillary", "generate_values"}:
                    continue
                elif "source" not in v.dtype.names:
                    raise ValueError("Data does not contain source information.")
            with self.assertRaises(
                TypeError, msg="Should raise error when directly instantiating StatisticalModel"
            ):
                model.data = data
                model.data["ancillary"] = None

    def test_likelihood(self):
        """Test of the _likelihood attribute."""
        for model, n in zip(self.models, self.n_likelihood_terms):
            # Check whether the likelihood is correctly set
            self.assertIsInstance(model._likelihood, LogLikelihoodSum)
            self.assertIsInstance(model.likelihood_list[-1], CustomAncillaryLikelihood)

            # Check length of likelihood
            self.assertEqual(len(model.likelihood_list), n + 1)

            # Check whether the likelihood is callable
            model.data = model.generate_data()
            model.ll()

    def test_fit(self):
        """Test of the fit method."""
        for model in self.models:
            model.data = model.generate_data()
            fit_result, max_llh = model.fit()

            # check whether all parameters are in fit_result
            self.assertEqual(set(model.parameters.names), set(fit_result.keys()))

            # check that non-fittable parameters are not fitted
            for p in model.parameters:
                if not p.fittable:
                    self.assertEqual(p.nominal_value, fit_result[p.name])

            # check that values are in fit limits
            for p in model.parameters:
                p.value_in_fit_limits(fit_result[p.name])

            # check that likelihood is maximized
            np.testing.assert_almost_equal(max_llh, model.ll(**fit_result), decimal=2)

            # check that fixing all parameters to nominal works
            fit_result_fixed, _ = model.fit(**model.parameters())
            for p in model.parameters:
                self.assertEqual(p.nominal_value, fit_result_fixed[p.name])

    def test_store_real_data(self):
        """Test of the store_real_data method."""
        for model, n in zip(self.models, self.n_likelihood_terms):
            data = model.generate_data()
            model.store_real_data(self.toydata_filename, list(data.values())[:n])
            toydata, _ = toydata_from_file(self.toydata_filename)
            model.data = toydata[0]
            model.fit()
            remove(self.toydata_filename)

    def test_needs_reinit(self):
        """Test of the needs_reinit property of parameters."""
        for model in self.models:
            self.assertTrue(model.parameters["wimp_mass"].needs_reinit)
            # check that a ValueError is raised when trying to set the nominal parameter
            with self.assertRaises(ValueError):
                model.parameters["wimp_mass"].nominal_value = 1
            # check that a ValueError is raised when calling the parameters with another value
            with self.assertRaises(ValueError):
                model.parameters(wimp_mass=1)
            with self.assertRaises(ValueError):
                model.get_expectation_values(wimp_mass=1)
            model.data = model.generate_data()
            with self.assertRaises(ValueError):
                model.fit(wimp_mass=1)

    def test_get_source_histograms(self):
        """Test of the get_source_histograms method."""
        for model in self.models:
            mus = model.get_expectation_values(per_likelihood_term=True)
            for ll_name, ll_term in zip(model.likelihood_names[:-1], model.likelihood_list[:-1]):
                source_histograms = model.get_source_histograms(ll_name)
                self.assertEqual(sorted(source_histograms.keys()), sorted(ll_term.source_name_list))

                # check whether the correct source histograms are returned
                for s_name, histogram in source_histograms.items():
                    source_index = ll_term.source_name_list.index(s_name)
                    blueice_source = ll_term.base_model.sources[source_index]
                    blueice_hist = blueice_source._pdf_histogram.histogram
                    np.testing.assert_almost_equal(blueice_hist, histogram, decimal=10)

                # check that expected_events boolean works
                source_histograms = model.get_source_histograms(ll_name, expected_events=True)
                for s_name, histogram in source_histograms.items():
                    mu = mus[ll_name][s_name]
                    sum_hist = histogram.n
                    np.testing.assert_almost_equal(mu, sum_hist, decimal=4)

            # check that model.likelihood_names[-1] fails
            with self.assertRaises(ValueError):
                model.get_source_histograms(model.likelihood_names[-1])

            # check that invalid likelihood names fail
            with self.assertRaises(ValueError):
                model.get_source_histograms("alea_iacta_est")

    def test_sorted_returns(self):
        """Test if sources are sorted in the same way for all return dicts."""
        for model in self.models:
            mus_per_ll = model.get_expectation_values(per_likelihood_term=True)
            mus = model.get_expectation_values()
            hist_per_ll = {}
            for ll_name in model.likelihood_names[:-1]:
                hist_per_ll[ll_name] = model.get_source_histograms(ll_name)
            # check that keys are the same for each SR
            for ll_name in model.likelihood_names[:-1]:
                self.assertEqual(mus_per_ll[ll_name].keys(), hist_per_ll[ll_name].keys())
            # check that global keys are the same
            all_keys = {v for d in mus_per_ll.values() for v in d.keys()}
            all_keys = sorted(all_keys)
            self.assertEqual(all_keys, sorted(mus.keys()))

    def test_apply_efficiency(self):
        """Test if efficiencies are properly applied."""
        model = self.models[0]
        nominal_n = model.get_expectation_values()["wimp"]
        for signal_eff in [0, 0.5, 1.0, 2.0]:
            model.data = model.generate_data()
            n = model.get_expectation_values(signal_efficiency=signal_eff)["wimp"]
            self.assertEqual(n, nominal_n * signal_eff)



from copy import deepcopy
from itertools import product
import logging

import numpy as np
import scipy.stats as sps
import multihist as mh
from blueice.likelihood import BinnedLogLikelihood, UnbinnedLogLikelihood

logging.basicConfig(level=logging.INFO)


class BlueiceDataGenerator:
    """A class for generating data from a blueice likelihood term.

    Attributes:
        ll: The blueice likelihood term.
        binned (bool): True if the likelihood term is binned.
        bincs (list): The bin centers of the likelihood term.
        direction_names (list): The names of the directions of the likelihood term.
        source_histograms (list): The histograms of the sources of the likelihood term.
        data_lengths (list): The number of bins of each component of the likelihood term.
        dtype (list): The data type of the likelihood term.
        last_kwargs (dict): The last kwargs used to generate data.
        mus: The expected number of events of each source of the likelihood term.
        parameters (list): The parameters of the likelihood term.
        ll_term (BinnedLogLikelihood or UnbinnedLogLikelihood): A blueice likelihood term.

    """

    def __init__(self, ll_term):
        """Initialize the BlueiceDataGenerator."""
        if isinstance(ll_term, BinnedLogLikelihood):
            self.binned = True
        elif isinstance(ll_term, UnbinnedLogLikelihood):
            self.binned = False
        else:
            raise NotImplementedError

        ll = deepcopy(ll_term)
        bincs = []  # bin centers of each component
        direction_names = []
        dtype = []
        data_lengths = []  # list of number of bins of each component
        analysis_space = ll_term.base_model.config["analysis_space"]

        source_histograms = {}
        for n in ll.source_name_list:
            source_histograms[n] = mh.Histdd(dimensions=analysis_space)

        for name, bin_edges in analysis_space:
            bincs.append(0.5 * (bin_edges[1:] + bin_edges[:-1]))
            dtype.append((name, float))
            data_lengths.append(len(bin_edges) - 1)
            direction_names.append(name)
        # IDEA: make source a string, not an int
        dtype.append(("source", int))

        data_binc = np.zeros(np.prod(data_lengths), dtype=dtype)
        for i, l in enumerate(product(*bincs)):
            for n, v in zip(direction_names, l):
                data_binc[n][i] = v

        ll.set_data(data_binc)

        self.ll = ll
        self.direction_names = direction_names
        self.source_histograms = source_histograms
        self.data_lengths = data_lengths
        self.dtype = dtype
        self.last_kwargs = {}
        self.first_call = True
        self.mus = ll.base_model.expected_events()
        self.parameters = list(ll.shape_parameters.keys())
        self.parameters += [n + "_rate_multiplier" for n in ll.rate_parameters.keys()]

    def simulate(self, filter_kwargs=True, n_toys=None, sample_n_toys=False, **kwargs):
        """Simulate toys for each source.

        Args:
            filter_kwargs (bool, optional (default=True)): If True,
                only parameters of the ll object are accepted as kwargs. Defaults to True.
            n_toys (int, optional (default=None)): If not None,
                a fixed number n_toys of toys is generated for each source component.
                Defaults to None.
            sample_n_toys (bool, optional (default=False)): If True,
                the number of toys is sampled from a Poisson distribution with mean n_toys.
                Defaults to False. Only works if n_toys is not None.

        Keyword Args:
            kwargs: The parameters pasted to the likelihood function.

        Returns:
            numpy.array: Array of simulated data for all sources in the given analysis space.
            The index "source" indicates the corresponding source of an entry.
            The dtype follows self.dtype.

        """
        self.compute_pdfs_and_mus(filter_kwargs=filter_kwargs, **kwargs)

        if n_toys is not None:
            if sample_n_toys:
                self.n_toys_rv = sps.poisson(n_toys).rvs()
                n_sources = np.full(len(self.ll.base_model.sources), self.n_toys_rv)
            else:
                n_sources = np.full(len(self.ll.base_model.sources), n_toys)
        else:
            # Generate a number n_source (according to the expectation value)
            # of toys for each source component:
            n_sources = sps.poisson(self.mus).rvs()
            if len(self.ll.base_model.sources) == 1:
                n_sources = np.array([n_sources])

        r_data = np.zeros(np.sum(n_sources), dtype=self.dtype)
        i_write = 0
        for i, n_source in enumerate(n_sources):
            if n_source > 0:  # dont generate if 0
                source_name = self.ll.source_name_list[i]
                rvs = self.source_histograms[source_name].get_random(n_source)
                for j, n in enumerate(self.direction_names):
                    r_data[n][i_write : i_write + n_source] = rvs[:, j]
                r_data["source"][i_write : i_write + n_source] = i
                i_write += n_source
        return r_data

    def compute_pdfs_and_mus(self, filter_kwargs=True, **kwargs) -> None:
        """Compute PDFs of the sources for the given parameters.

        Args:
            filter_kwargs (bool, optional (default=True)): If True,
                only parameters of the ll object are accepted as kwargs. Defaults to True.
            kwargs: The parameters pasted to the likelihood function.

        """
        if filter_kwargs:
            kwargs = {k: v for k, v in kwargs.items() if k in self.parameters + ["livetime_days"]}

        # check if the cached generator may be used:
        unmatched_item = set(self.last_kwargs.items()) ^ set(kwargs.items())
        kwargs_changed = len(unmatched_item) != 0
        if self.first_call or kwargs_changed:
            ret = self.ll(full_output=True, **kwargs)
            if isinstance(ret, float):
                logging.warning("ERROR, generator kwarg outside range?")
                logging.warning(kwargs)
            _, mus, ps_array = ret
            for n, p in zip(self.ll.source_name_list, ps_array):
                self.source_histograms[n].histogram = p.reshape(self.data_lengths)
                if not self.binned:
                    self.source_histograms[n] *= self.source_histograms[n].bin_volumes()
            self.mus = mus
            self.last_kwargs = kwargs
            self.first_call = False



from os import remove
import inspect
import pytest
from unittest import TestCase

import numpy as np
from inference_interface import toyfiles_to_numpy

from alea.utils import load_yaml
from alea.runner import Runner

from .test_gaussian_model import gaussian_model_parameter_definition


COMPUTE_CONFIDENCE_INTERVAL = True


@pytest.mark.usefixtures("rm_cache")
class TestRunner(TestCase):
    """Test of the Runner class."""

    @classmethod
    def setUp(cls):
        """Initialise the Runner instance."""
        cls.running_config = load_yaml("unbinned_wimp_running.yaml")
        cls.toydata_filename = "simple_data.ii.h5"
        cls.output_filename = "test_toymc.ii.h5"
        cls.n_mc = 3

    def set_gaussian_runner(self, toydata_mode="generate_and_store"):
        """Set a new runner instance with GaussianModel."""
        self.runner = Runner(
            statistical_model="alea.examples.gaussian_model.GaussianModel",
            poi="mu",
            hypotheses=["free", "zero", "true"],
            n_mc=self.n_mc,
            generate_values={"mu": 1.0},
            nominal_values={"sigma": 1.0},
            parameter_definition=gaussian_model_parameter_definition,
            compute_confidence_interval=COMPUTE_CONFIDENCE_INTERVAL,
            toydata_mode=toydata_mode,
            toydata_filename=self.toydata_filename,
            output_filename=self.output_filename,
        )

    def set_blueice_runner(self, toydata_mode="generate_and_store"):
        """Set a new runner instance with BlueiceExtendedModel."""
        parameter_zvc = self.running_config["computation_options"]["discovery_power"]
        self.runner = Runner(
            statistical_model=self.running_config["statistical_model"],
            poi=self.running_config["poi"],
            hypotheses=parameter_zvc["in_common"]["hypotheses"],
            n_mc=self.n_mc,
            generate_values={"wimp_rate_multiplier": 1.0},
            statistical_model_config=self.running_config["statistical_model_config"],
            compute_confidence_interval=COMPUTE_CONFIDENCE_INTERVAL,
            toydata_mode=toydata_mode,
            toydata_filename=self.toydata_filename,
            output_filename=self.output_filename,
        )

    def test_runners(self):
        """Test of the simulate_and_fit and write_output method."""
        set_runners = [self.set_gaussian_runner, self.set_blueice_runner]
        for set_runner in set_runners:
            # test toydata_mode generate_and_store
            set_runner()
            self.runner.run()
            remove(self.output_filename)

            # test toydata_mode read
            set_runner(toydata_mode="read")
            self.runner.run()
            remove(self.toydata_filename)

            # check confidence interval computation
            if COMPUTE_CONFIDENCE_INTERVAL:
                results = toyfiles_to_numpy(self.runner._output_filename)
                mask = np.any(np.isnan(results["free"]["dl"]))
                mask &= np.any(np.isnan(results["free"]["ul"]))
                if mask:
                    raise ValueError("Confidence interval computation failed!")
            remove(self.output_filename)

    def test_init_signatures(self):
        """Test the signatures of the Runner.__init__"""
        (
            args,
            varargs,
            varkw,
            defaults,
            kwonlyargs,
            kwonlydefaults,
            annotations,
        ) = inspect.getfullargspec(Runner.__init__)
        if (len(annotations) != len(args[1:])) or (len(defaults) != len(args[1:])):
            raise ValueError(
                "The number of annotations and defaults of Runner.__init__ must be the same!"
            )



from unittest import TestCase

from alea.models import BlueiceExtendedModel
from alea.utils import load_yaml


class TestTemplateSource(TestCase):
    """Test of the TemplateSource class."""

    def test_init_templates(self):
        """Test whether we can initialize template sources."""
        model_configs = load_yaml("unbinned_wimp_statistical_model_template_source_test.yaml")
        parameter_definition = model_configs["parameter_definition"]
        likelihood_config = model_configs["likelihood_config"]
        model = BlueiceExtendedModel(parameter_definition, likelihood_config)
        model.nominal_expectation_values

    def test_wrong_analysis_space(self):
        """Test whether initializing with a wrong analysis_space raises error."""
        model_configs = load_yaml("unbinned_wimp_statistical_model_template_source_test.yaml")
        parameter_definition = model_configs["parameter_definition"]
        likelihood_config = model_configs["likelihood_config"]
        # Change the analysis space to a wrong one
        space = likelihood_config["likelihood_terms"][0]["analysis_space"]
        # additive mismatch in cs1
        space[0]["cs1"] = "np.linspace(1, 101, 51)"
        space[1]["cs2"] = "np.geomspace(100, 100000, 51)"
        with self.assertRaises(AssertionError):
            _ = BlueiceExtendedModel(parameter_definition, likelihood_config)

        # multiplicative mismatch in cs2
        space[0]["cs1"] = "np.linspace(0, 100, 51)"
        space[1]["cs2"] = "np.geomspace(101, 101000, 51)"
        with self.assertRaises(AssertionError):
            _ = BlueiceExtendedModel(parameter_definition, likelihood_config)

        # If the dimensions are wrong we should get ValueError
        space[0]["cs1"] = "np.linspace(0, 100, 50)"
        space[1]["cs2"] = "np.geomspace(100, 100000, 51)"
        with self.assertRaises(ValueError):
            _ = BlueiceExtendedModel(parameter_definition, likelihood_config)



import warnings
from typing import Any, Dict, List, Tuple, Iterator, Optional, Union, cast
import pandas as pd

from alea.utils import within_limits, clip_limits, evaluate_numpy_scipy_expression


class Parameter:
    """Represents a single parameter with its properties.

    Attributes:
        name (str): The name of the parameter.
        nominal_value (float, optional (default=None)): The nominal value of the parameter.
        fittable (bool, optional (default=True)):
            Indicates if the parameter is fittable or always fixed.
        ptype (str, optional (default=None)): The ptype of the parameter.
        uncertainty (float or str, optional (default=None)): The uncertainty of the parameter.
            If a string, it can be evaluated as a numpy or
            scipy function to define non-gaussian constraints.
        relative_uncertainty (bool, optional (default=None)):
            Indicates if the uncertainty is relative to the nominal_value.
        blueice_anchors (list, optional (default=None)): Anchors for blueice template morphing.
            Blueice will load the template for the provided values and then interpolate
            for any value in between.
        fit_limits (Tuple[float, float], optional (default=None)):
            The limits for fitting the parameter.
        parameter_interval_bounds (Tuple[float, float], optional (default=None)):
            Limits for computing confidence intervals.
        fit_guess (float, optional (default=None)): The initial guess for fitting the parameter.
        description (str, optional (default=None)): A description of the parameter.

    """

    _uncertainty: Optional[Union[float, str]]

    def __init__(
        self,
        name: str,
        nominal_value: Optional[float] = None,
        fittable: bool = True,
        ptype: Optional[str] = None,
        uncertainty: Optional[Union[float, str]] = None,
        relative_uncertainty: Optional[bool] = None,
        blueice_anchors: Optional[Union[list, str]] = None,
        fit_limits: Optional[Tuple] = None,
        parameter_interval_bounds: Optional[Tuple[float, float]] = None,
        fit_guess: Optional[float] = None,
        description: Optional[str] = None,
    ):
        """Initialise a parameter."""
        self.name = name
        self._nominal_value = nominal_value
        self.fittable = fittable
        self.ptype = ptype
        self.relative_uncertainty = relative_uncertainty
        self.uncertainty = uncertainty
        self.blueice_anchors = blueice_anchors
        self.fit_limits = fit_limits
        self.parameter_interval_bounds = parameter_interval_bounds
        self.fit_guess = fit_guess
        self.description = description

        self._check_parameter_consistency()

    def __repr__(self) -> str:
        parameter_str = ", ".join([f"{k}={v}" for k, v in self.__dict__.items() if v is not None])
        _repr = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        _repr += f"({parameter_str})"
        return _repr

    @property
    def uncertainty(self) -> Any:
        """Return the uncertainty of the parameter.

        If the uncertainty is a string, it will be evaluated as a numpy or scipy function.

        """
        if isinstance(self._uncertainty, str):
            return evaluate_numpy_scipy_expression(self._uncertainty)
        else:
            return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, value: Optional[Union[float, str]]) -> None:
        if self.relative_uncertainty and (value is not None):
            if value and (not isinstance(value, (float, int))):
                raise ValueError(
                    f"When relative_uncertainty of {self.name} is True, "
                    f"uncertainty should be float, not {value}."
                )
            if self.nominal_value is None:
                raise ValueError(
                    f"When relative_uncertainty of {self.name} is True, "
                    "nominal_value should be set."
                )
        self._uncertainty = value

    @property
    def blueice_anchors(self) -> Any:
        """Return the blueice_anchors of the parameter.

        If the blueice_anchors is a string, it will be evaluated as a numpy or scipy function.

        """
        if isinstance(self._blueice_anchors, str):
            return evaluate_numpy_scipy_expression(self._blueice_anchors).tolist()
        else:
            return self._blueice_anchors

    @blueice_anchors.setter
    def blueice_anchors(self, value: Optional[Union[list, str]]) -> None:
        self._blueice_anchors = value

    @property
    def fit_guess(self) -> Optional[float]:
        """Return the initial guess for fitting the parameter."""
        # make sure to only return fit_guess if fittable
        if self._fit_guess is not None and not self.fittable:
            raise ValueError(f"Parameter {self.name} is not fittable, but has a fit_guess.")
        else:
            return self._fit_guess

    @fit_guess.setter
    def fit_guess(self, value: Optional[float]) -> None:
        self._fit_guess = value

    @property
    def parameter_interval_bounds(self) -> Optional[Tuple[float, float]]:
        # make sure to only return parameter_interval_bounds if fittable
        if self._parameter_interval_bounds is not None and not self.fittable:
            raise ValueError(
                f"Parameter {self.name} is not fittable, but has a parameter_interval_bounds."
            )
        else:
            # print warning when value contains None
            value = self._parameter_interval_bounds
            self._check_parameter_interval_bounds(value)
            return clip_limits(value)

    @parameter_interval_bounds.setter
    def parameter_interval_bounds(self, value: Optional[Tuple[float, float]]) -> None:
        self._parameter_interval_bounds = value

    @property
    def nominal_value(self) -> Optional[float]:
        """Return the nominal value of the parameter."""
        return self._nominal_value

    @nominal_value.setter
    def nominal_value(self, value: Optional[float]) -> None:
        if self.needs_reinit and (value != self._nominal_value):
            raise ValueError(
                f"{self.name} is a parameter that requires re-initialization "
                "to change its nominal value "
                f"(tried to override nominal value {self._nominal_value} with {value})."
            )
        self._nominal_value = value

    @property
    def needs_reinit(self) -> bool:
        """Return True if the parameter needs re-initialization (for ptype ``needs_reinit``)."""
        needs_reinit = False
        if self.ptype == "needs_reinit":
            needs_reinit = True
        return needs_reinit

    def __eq__(self, other: object) -> bool:
        """Return True if all attributes are equal."""
        if isinstance(other, Parameter):
            return all(getattr(self, k) == getattr(other, k) for k in self.__dict__)
        else:
            return False

    def value_in_fit_limits(self, value: float) -> bool:
        """Returns True if value is within fit_limits."""
        return within_limits(value, self.fit_limits)

    def _check_parameter_interval_bounds(self, value):
        """Check if parameter_interval_bounds is within fit_limits and is not None."""
        if (value is None) or (value[0] is None) or (value[1] is None):
            warnings.warn(
                f"parameter_interval_bounds not completely defined for parameter {self.name}. "
                "This may cause numerical overflow when calculating confidential interval."
            )
        value = clip_limits(value)
        if not (self.value_in_fit_limits(value[0]) and self.value_in_fit_limits(value[1])):
            raise ValueError(
                f"parameter_interval_bounds {value} not within "
                f"fit_limits {self.fit_limits} for parameter {self.name}."
            )

    def _check_parameter_consistency(self):
        """Check if parameter is consistent."""
        if self.fittable and self.needs_reinit:
            warnings.warn(
                f"Parameter {self.name} is fittable and needs re-initialization. "
                "This may cause unexpected behaviour."
            )
        if (self.blueice_anchors is not None) and self.needs_reinit:
            raise ValueError(
                f"Parameter {self.name} needs re-initialization but has "
                "blueice_anchors defined. "
                "This may cause unexpected behaviour."
            )


class ConditionalParameter:
    """This class is used to define a parameter that depends on another parameter. It has the same
    attributes as the Parameter class but each of them can be a dictionary with keys being the
    values of the conditioning parameter and values being the corresponding values of the
    conditional parameter. Calling the object with the conditioning parameter value as an argument
    will return a corresponding Parameter object with the correct values.

    Attributes:
        name (str): The name of the parameter.
        conditioning_parameter_name (str): The name of the conditioning parameter.

    """

    def __init__(self, name: str, conditioning_parameter_name: str, **kwargs):
        self.name = name
        self.conditioning_name = conditioning_parameter_name
        self.conditions_dict = self._unpack_conditions(kwargs)
        self.conditioning_param = None

    def __repr__(self) -> str:
        parameter_str = ", ".join([f"{k}={v}" for k, v in self.__dict__.items() if v is not None])
        _repr = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        _repr += f"({parameter_str})"
        return _repr

    @staticmethod
    def _unpack_conditions(kwargs):
        # 1) collect all condition keys and check for consistency
        all_keys = set()
        for value in kwargs.values():
            if isinstance(value, dict):
                if not all_keys:
                    all_keys = set(value.keys())
                elif all_keys != set(value.keys()):
                    raise ValueError("Inconsistent condition keys across dictionaries.")

        # 2) create the conditions dictionary
        conditions_dict = {key: {} for key in all_keys}
        for key, value in kwargs.items():
            if isinstance(value, dict):
                for condition_key, condition_value in value.items():
                    conditions_dict[condition_key][key] = condition_value
            else:
                for condition_key in all_keys:
                    conditions_dict[condition_key][key] = value

        return conditions_dict

    @property
    def uncertainty(self) -> Any:
        """Return the uncertainty of the parameter (cominal condition)"""
        return self().uncertainty

    @property
    def blueice_anchors(self) -> Any:
        """Return the blueice_anchors of the parameter (cominal condition)"""
        return self().blueice_anchors

    @property
    def fit_guess(self) -> Optional[float]:
        """Return the initial guess for fitting the parameter (cominal condition)"""
        return self().fit_guess

    @property
    def parameter_interval_bounds(self) -> Optional[Tuple[float, float]]:
        """Return the parameter_interval_bounds of the parameter (cominal condition)"""
        return self().parameter_interval_bounds

    @property
    def nominal_value(self) -> Optional[float]:
        """Return the nominal value of the parameter (cominal condition)"""
        return self().nominal_value

    @property
    def needs_reinit(self) -> bool:
        """Return True if the parameter needs re-initialization (for ptype ``needs_reinit``)."""
        return self().needs_reinit

    @property
    def fittable(self) -> bool:
        """Return the fittable attribute of the parameter (cominal condition)"""
        return self().fittable

    @property
    def ptype(self) -> Optional[str]:
        """Return the ptype of the parameter (cominal condition)"""
        return self().ptype

    @property
    def relative_uncertainty(self) -> Optional[bool]:
        """Return the relative_uncertainty of the parameter (cominal condition)"""
        return self().relative_uncertainty

    @property
    def fit_limits(self) -> Optional[Tuple[float, float]]:
        """Return the fit_limits of the parameter (cominal condition)"""
        return self().fit_limits

    def __eq__(self, other: object) -> bool:
        """Return True if all attributes are equal."""
        if isinstance(other, ConditionalParameter):
            return all(getattr(self, k) == getattr(other, k) for k in self.__dict__)
        return False

    def value_in_fit_limits(self, value: float) -> bool:
        """Returns True if value under cominal condition is within fit_limits."""
        return self().value_in_fit_limits(value)

    def __call__(self, **kwargs) -> Parameter:
        if self.conditioning_name in kwargs:
            cond_val = kwargs[self.conditioning_name]
        elif self.conditioning_param is not None:
            cond_val = self.conditioning_param.nominal_value
        else:
            err_msg = (
                f"Conditioning parameter '{self.conditioning_name}' is missing. Can't fall back to "
                "nominal value because conditioning parameter it is not set. "
            )
            raise ValueError(err_msg)
        # check if the conditioning value is in the conditions dictionary
        if cond_val not in self.conditions_dict:
            raise ValueError(
                f"Conditioning value '{cond_val}' not found in the conditions dictionary."
                + f"Available values are: {sorted(list(self.conditions_dict.keys()))}"
            )
        return Parameter(name=self.name, **self.conditions_dict[cond_val])


class Parameters:
    """Represents a collection of parameters.

    Attributes:
        names (List[str]): A list of parameter names.
        fit_guesses (Dict[str, float]): A dictionary of fit guesses.
        fit_limits (Dict[str, float]): A dictionary of fit limits.
        fittable (List[str]): A list of parameter names which are fittable.
        not_fittable (List[str]): A list of parameter names which are not fittable.
        uncertainties (Dict[str, float or Any]): A dictionary of parameter uncertainties.
        with_uncertainty (Parameters): A Parameters object with parameters with
            a not-NaN uncertainty.
        nominal_values (Dict[str, float]): A dictionary of parameter nominal values.
        parameters (Dict[str, Parameter]): A dictionary to store the parameters,
            with parameter name as key.

    """

    def __init__(self):
        """Initialise a collection of parameters."""
        self.parameters = cast(Dict[str, Parameter], {})

    def __iter__(self) -> Iterator[Parameter]:
        """Return an iterator over the parameters.

        Each iteration return a Parameter object.

        """
        return iter(self.parameters.values())

    @classmethod
    def from_config(cls, config: Dict[str, dict]):
        """Creates a Parameters object from a configuration dictionary.

        Args:
            config (dict): A dictionary of parameter configurations.

        Returns:
            Parameters: The created Parameters object.

        """
        parameters = cls()
        parameter: Union[Parameter, ConditionalParameter]
        for name, param_config in config.items():
            if "conditioning_parameter_name" in param_config:
                parameter = ConditionalParameter(name, **param_config)
            else:
                parameter = Parameter(name=name, **param_config)
            parameters.add_parameter(parameter)
        # set conditioning parameters
        for param in parameters:
            if isinstance(param, ConditionalParameter):
                param.conditioning_param = parameters[param.conditioning_name]
        return parameters

    @classmethod
    def from_list(cls, names: List[str]):
        """Creates a Parameters object from a list of parameter names. Everything else is set to
        default values.

        Args:
            names (List[str]): List of parameter names.

        Returns:
            Parameters: The created Parameters object.

        """
        parameters = cls()
        for name in names:
            parameter = Parameter(name)
            parameters.add_parameter(parameter)
        return parameters

    def __repr__(self) -> str:
        parameter_str = ", ".join(self.names)
        _repr = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        _repr += f"({parameter_str})"
        return _repr

    def __str__(self) -> str:
        """Return an overview table of all parameters."""
        par_list = []
        for p in self:
            if isinstance(p, ConditionalParameter):
                par_dict = {
                    "conditioning_name": p.conditioning_name,
                    "conditions": sorted(p.conditions_dict.keys()),
                }
                # get nominal-condition parameter
                p = p()
            else:
                par_dict = {}
            for k, v in p.__dict__.items():
                # replace hidden attributes with non-hidden properties
                if k.startswith("_"):
                    par_dict[k[1:]] = v
                else:
                    par_dict[k] = v
            par_list.append(par_dict)

        df = pd.DataFrame(par_list)
        # make name column the index
        df.set_index("name", inplace=True)
        df.index.name = None

        return df.to_string()

    def add_parameter(self, parameter: Union[Parameter, ConditionalParameter]) -> None:
        """Adds a Parameter object to the Parameters collection.

        Args:
            parameter (Parameter): The Parameter object to add.

        Raises:
            ValueError: If the parameter name already exists.

        """
        if parameter.name in self.names:
            raise ValueError(f"Parameter {parameter.name} already exists.")
        self.parameters[parameter.name] = parameter

    @property
    def names(self) -> List[str]:
        """A list of parameter names."""
        return list(self.parameters.keys())

    @property
    def fit_guesses(self) -> Dict[str, float]:
        """A dictionary of fit guesses."""
        return {
            name: param.fit_guess
            for name, param in self.parameters.items()
            if param.fit_guess is not None
        }

    @property
    def fit_limits(self) -> Dict[str, float]:
        """A dictionary of fit limits."""
        return {
            name: param.fit_limits
            for name, param in self.parameters.items()
            if param.fit_limits is not None
        }

    @property
    def fittable(self) -> List[str]:
        """A list of parameter names which are fittable."""
        return [name for name, param in self.parameters.items() if param.fittable]

    @property
    def not_fittable(self) -> List[str]:
        """A list of parameter names which are not fittable."""
        return [name for name, param in self.parameters.items() if not param.fittable]

    @property
    def uncertainties(self) -> dict:
        """A dict of uncertainties for all parameters with a not-NaN uncertainty.

        Caution: this is not the same as the parameter.uncertainty property.

        """
        return {k: i.uncertainty for k, i in self.parameters.items() if i.uncertainty is not None}

    @property
    def with_uncertainty(self) -> "Parameters":
        """Return parameters with a not-NaN uncertainty.

        The parameters are the same objects as in the original Parameters object, not a copy. For
        conditional parameters, the parameters under the nominal condition are returned.

        """
        param_dict = {k: i for k, i in self.parameters.items() if i.uncertainty is not None}
        params = Parameters()
        for param in param_dict.values():
            params.add_parameter(param)
        return params

    @property
    def nominal_values(self) -> dict:
        """A dict of nominal values for all parameters with a nominal value."""
        return {
            k: i.nominal_value for k, i in self.parameters.items() if i.nominal_value is not None
        }

    def set_nominal_values(self, **nominal_values):
        """Set the nominal values for parameters.

        Keyword Args:
            nominal_values (dict): A dict of parameter names and values.

        """
        for name, value in nominal_values.items():
            self.parameters[name].nominal_value = value

    def set_fit_guesses(self, **fit_guesses):
        """Set the fit guesses for parameters.

        Keyword Args:
            fit_guesses (dict): A dict of parameter names and values.

        """
        for name, value in fit_guesses.items():
            self.parameters[name].fit_guess = value

    def _evaluate_parameter(self, parameter: Parameter, **kwargs):
        if isinstance(parameter, ConditionalParameter):
            return parameter(**kwargs)
        return parameter

    def __call__(
        self, return_fittable: Optional[bool] = False, **kwargs: Optional[Dict]
    ) -> Dict[str, float]:
        """Return a dictionary of parameter values, optionally filtered to return only fittable
        parameters.

        Args:
            return_fittable (bool, optional (default=False)):
                Indicates if only fittable parameters should be returned.

        Keyword Args:
            kwargs (dict): Additional keyword arguments to override parameter values.

        Raises:
            ValueError: If a parameter name is not found.

        Returns:
            dict: A dictionary of parameter values.

        """
        values = {}

        # check that all kwargs are valid parameter names
        for name in kwargs:
            if name not in self.parameters:
                raise ValueError(f"Parameter '{name}' not found.")

        for name, param in self.parameters.items():
            param = self._evaluate_parameter(param, **kwargs)
            new_val = kwargs.get(name, None)
            if param.needs_reinit and new_val != param.nominal_value and new_val is not None:
                raise ValueError(
                    f"{name} is a parameter that requires re-initialization "
                    "to override its nominal value "
                    f"(tried to override nominal value {param.nominal_value} "
                    f"with {new_val})."
                )
            if (return_fittable and param.fittable) or (not return_fittable):
                values[name] = new_val if new_val is not None else param.nominal_value
        if any(i is None for k, i in values.items()):
            emptypars = ", ".join([k for k, i in values.items() if i is None])
            raise AssertionError(
                "All parameters must be set explicitly, or have a nominal value, "
                "not satisfied for: " + emptypars
            )
        return values

    def __getattr__(self, name: str) -> Parameter:
        """Retrieves a Parameter object by attribute access.

        Args:
            name (str): The name of the parameter.

        Raises:
            AttributeError: If the attribute is not found.

        Returns:
            Parameter: The retrieved Parameter object.

        """
        try:
            return super().__getattribute__("parameters")[name]
        except KeyError:
            raise AttributeError(f"Attribute '{name}' not found.")

    def __getitem__(self, name: str) -> Parameter:
        """Retrieves a Parameter object by dictionary access.

        Args:
            name (str): The name of the parameter.

        Raises:
            KeyError: If the key is not found.

        Returns:
            Parameter: The retrieved Parameter object.

        """
        if name in self.parameters:
            return self.parameters[name]
        else:
            raise KeyError(f"Key '{name}' not found.")

    def __eq__(self, other: object) -> bool:
        """Return True if all parameters are equal."""
        if isinstance(other, Parameters):
            names = set(self.names + other.names)
            return all(getattr(self, n) == getattr(other, n) for n in names)
        else:
            return False

    def values_in_fit_limits(self, **kwargs: Dict) -> bool:
        """Return True if all values are within the fit limits.

        Keyword Args:
            kwargs (dict): The parameter values to check.

        Returns:
            bool: True if all values are within the fit limits.

        """
        for name, value in kwargs.items():
            param = self._evaluate_parameter(self.parameters[name], **kwargs)
            if not param.value_in_fit_limits(value):
                return False
        return True
